import numpy as np
# from pandas import read_csv
from aston.trace import Chromatogram
from aston.tracefile import TraceFile


class CSVFile(TraceFile):
    '''
    Reads in a *.CSV. Assumes that the first line is the header and
    that the file is comma delimited.
    '''
    mime = 'text/csv'
    traces = ['#']
    # TODO: use pandas to make this much better?
    # TODO: determine traces to list

    @property
    def data(self):
        delim = ','
        try:  # TODO: better, smarter error checking than this
            with open(self.filename, 'r') as f:
                lns = f.readlines()
                ions = [float(i) for i in lns[0].split(delim)[1:]]
                data = np.array([np.fromstring(ln, sep=delim)
                                 for ln in lns[1:]])
                return Chromatogram(data[:, 1:], data[:, 0], ions)
        except Exception:
            return Chromatogram()


def write_csv(filename, df, info=None, delimiter=','):
    with open(filename, 'w') as f:
        # write out columns
        f.write('Time' + delimiter)
        f.write(delimiter.join(str(i) for i in df.data.columns.tolist()))

        # write out the rest, scan by scan
        for scan in df.scans():
            f.write('\n ' + str(df.name) + delimiter)
            f.write(delimiter.join(str(i) for i in scan.abn.tolist()))



import numpy as np
from scipy.optimize import leastsq, fmin, fmin_l_bfgs_b

# bounding code inspired by http://newville.github.com/lmfit-py/bounds.html
# which was inspired by leastsqbound, which was inspired by MINUIT


def _to_bound_p(params, bounds):
    new_v = {}
    for p in params:
        if p not in bounds:
            new_v[p] = params[p]
            continue
        b = bounds[p]
        if b[1] == np.inf and b[0] == -np.inf:
            # already in the correct support: ignore
            new_v[p] = params[p]
        elif b[1] == np.inf:
            new_v[p] = b[0] - 1. + np.sqrt(params[p] ** 2 + 1.)
        elif b[0] == -np.inf:
            new_v[p] = b[1] + 1. - np.sqrt(params[p] ** 2 + 1.)
        else:
            new_v[p] = b[0] + 0.5 * np.sin(params[p] + 1.) * (b[1] - b[0])
    return new_v


def _to_unbnd_p(params, bounds):
    new_v = {}
    for p in params:
        if p not in bounds:
            new_v[p] = params[p]
            continue
        b = bounds[p]
        if (b[1] == np.inf and b[0] == -np.inf):
            # already in the correct support: ignore
            new_v[p] = params[p]
        elif b[1] == np.inf:
            new_v[p] = np.sqrt((params[p] - b[0] + 1.) ** 2 - 1.)
        elif b[0] == -np.inf:
            new_v[p] = np.sqrt((b[1] - params[p] + 1.) ** 2 - 1.)
        else:
            new_v[p] = np.arcsin(2 * (params[p] - b[0]) / (b[1] - b[0]) - 1)
    return new_v


def guess_initc(ts, f, rts=[]):
    """
    ts - An AstonSeries that's being fitted with peaks
    f - The functional form of the peaks (e.g. gaussian)
    rts - peak maxima to fit; each number corresponds to one peak
    """
    def find_side(y, loc=None):
        if loc is None:
            loc = y.argmax()
        ddy = np.diff(np.diff(y))
        lft_loc, rgt_loc = loc - 2, loc + 1
        while rgt_loc >= 0 and rgt_loc < len(ddy):
            if ddy[rgt_loc] < ddy[rgt_loc - 1]:
                break
            rgt_loc += 1
        while lft_loc >= 0 and lft_loc < len(ddy):
            if ddy[lft_loc] < ddy[lft_loc + 1]:
                break
            lft_loc -= 1
        return lft_loc + 1, rgt_loc + 1

    # weight_mom = lambda m, a, w: \
    #   np.sum(w * (a - np.sum(w * a) / np.sum(w)) ** m) / np.sum(w)
    # sig = np.sqrt(weight_mom(2, ts.index, ts.values))  # sigma
    # peak_params['s'] = weight_mom(3, ts.index, ts.values) / sig ** 3
    # peak_params['e'] = weight_mom(4, ts.index, ts.values) / sig ** 4 - 3
    # TODO: better method of calculation of these?
    all_params = []
    for rt in rts:
        peak_params = {'x': rt}  # ts.index[ts.values.argmax()]
        top_idx = np.abs(ts.index - rt).argmin()
        side_idx = find_side(ts.values, top_idx)
        peak_params['h'] = ts.values[top_idx]
        # - min(ts.y[side_idx[0]], ts.y[side_idx[1]])
        peak_params['w'] = ts.index[side_idx[1]] - ts.index[side_idx[0]]
        peak_params['s'] = 1.1
        peak_params['e'] = 1.
        peak_params['a'] = 1.
        all_params.append(peak_params)
    return all_params


def fit(ts, fs=[], all_params=[], fit_vars=None,
        alg='leastsq', make_bounded=True):
    """
    Use a minimization algorithm to fit a AstonSeries with
    analytical functions.
    """
    if fit_vars is None:
        fit_vars = [f._peakargs for f in fs]
    initc = [min(ts.values)]
    for f, peak_params, to_fit in zip(fs, all_params, fit_vars):
        if 'v' in to_fit:
            to_fit.remove('v')

        if make_bounded and hasattr(f, '_pbounds'):
            new_v = _to_unbnd_p({i: peak_params[i] for i in to_fit},
                                f._pbounds)
            initc += [new_v[i] for i in to_fit]
        else:
            initc += [peak_params[i] for i in to_fit]

    def errfunc_lsq(fit_params, t, y, all_params):
        # first value in fit_params is baseline
        # fit_y = np.ones(len(t)) * fit_params[0]
        fit_y = np.zeros(len(t))
        param_i = 1
        for f, peak_params, to_fit in zip(fs, all_params, fit_vars):
            for k in to_fit:
                peak_params[k] = fit_params[param_i]
                param_i += 1
            if make_bounded and hasattr(f, '_pbounds'):
                fit_y += f(t, **_to_bound_p(peak_params, f._pbounds))
            else:
                fit_y += f(t, **peak_params)
        return fit_y - y

    def errfunc(p, t, y, all_params):
        return np.sum(errfunc_lsq(p, t, y, all_params) ** 2)

    if alg == 'simplex':
        fit_p, _ = fmin(errfunc, initc, args=(ts.index, ts.values,
                                              peak_params))
#    elif alg == 'anneal':
#        fit_p, _ = anneal(errfunc, initc, args=(ts.index, ts.values,
#                                                peak_params))
    elif alg == 'lbfgsb':
        # TODO: use bounds param
        fitp, _ = fmin_l_bfgs_b(errfunc, fit_p,
                                args=(ts.index, ts.values, peak_params),
                                approx_grad=True)
    elif alg == 'leastsq':
        fit_p, _ = leastsq(errfunc_lsq, initc, args=(ts.index, ts.values,
                                                     all_params))
    # else:
    #     r = minimize(errfunc, initc, \
    #                  args=(ts.index, ts.values, all_params), \
    #                  jac=False, gtol=1e-2)
    #     #if not r['success']:
    #     #    print('Fail:' + str(f))
    #     #    print(r)
    #     #if np.nan in r['x']:  # not r['success']?
    #     #    fit_p = initc
    #     #else:
    #     #    fit_p = r['x']

    fit_pl = fit_p.tolist()
    v = fit_pl.pop(0)  # noqa
    fitted_params = []
    for f, to_fit in zip(fs, fit_vars):
        fit_p_dict = {v: fit_pl.pop(0) for v in to_fit}
        # fit_p_dict['v'] = v
        if make_bounded and hasattr(f, '_pbounds'):
            fitted_params.append(_to_bound_p(fit_p_dict, f._pbounds))
        else:
            fitted_params.append(fit_p_dict)

    # calculate r^2 of the fit
    ss_err = errfunc(fit_p, ts.index, ts.values, fitted_params)
    ss_tot = np.sum((ts.values - np.mean(ts.values)) ** 2)
    r2 = 1 - ss_err / ss_tot
    res = {'r^2': r2}

    return fitted_params, res



import re
import zlib
import base64
from xml.etree import ElementTree as Et
import numpy as np
from aston import __version__
from aston.trace import Trace
from aston.tracefile import ScanListFile
from aston.spectra import Scan


def t_to_min(x):
    """
    Convert XML 'xs: duration type' to decimal minutes, e.g.:
    t_to_min('PT1H2M30S') == 62.5
    """
    g = re.match('PT(?:(.*)H)?(?:(.*)M)?(?:(.*)S)?', x).groups()
    return sum(0 if g[i] is None else float(g[i]) * 60. ** (1 - i)
               for i in range(3))


# class mzXML(ScanListFile):
class MzXML(object):
    mime = 'application/mzxml'
    traces = ['#ms']

    ns = {'m': 'http://sashimi.sourceforge.net/schema_revision/mzXML_2.1'}

    @property
    def data(self):
        # get all the scans under the root node
        r = Et.parse(self.filename).getroot()
        s = r.findall('*//m:scan', namespaces=self.ns)

        for scn in s:
            # scn.get('peaksCount')
            # extract out the actual scan data
            pks = scn.find('m:peaks', namespaces=self.ns)
            dtype = {'32': '>f4', '64': '>f8'}.get(pks.get('precision'))
            d = np.frombuffer(base64.b64decode(pks.text), dtype)
            mz = d[::2]  # noqa
            abn = d[1::2]  # noqa
            # FIXME

    def total_trace(self, twin=None):
        # TODO: use twin
        # TODO: only get the scans with totIonCurrent; if none found
        # calculate from the data

        r = Et.parse(self.filename).getroot()
        s = r.findall('*//m:scan', namespaces=self.ns)
        d = np.array([float(i.get('totIonCurrent')) for i in s])
        t = np.array([t_to_min(i.get('retentionTime')) for i in s])
        return Trace(d, t, name='TIC')


class MzML(ScanListFile):
    mime = 'application/mzml'
    traces = ['#ms']

    ns = {'m': 'http://psi.hupo.org/ms/mzml'}

    def scans(self, twin=None):
        if twin is None:
            twin = (-np.inf, np.inf)
        r = Et.parse(self.filename).getroot()
        pgr = r.find('.//m:referenceableParamGroupList', namespaces=self.ns)

        spectra = r.findall('*//m:spectrum', namespaces=self.ns)
        for s in spectra:
            q = './/m:cvParam[@accession="MS:1000016"]'
            time_elem = s.find(q, namespaces=self.ns)
            if time_elem is None:
                continue
            time = time_elem.get('value')

            # FIXME: won't find these properties if a paramGroupRef exists
            for i in ['MS:1000514', 'MS:1000617', 'MS:1000786']:
                q = './/m:cvParam[@accession="' + i + '"]/..'
                x_elem = s.find(q, namespaces=self.ns)
                if x_elem is not None:
                    x = self.read_binary(x_elem, pgr)
                    break
            else:
                # check paramGroupRef
                q0 = './/m:referenceableParamGroupRef'
                bin_arrs = s.findall(q0 + '/..', namespaces=self.ns)
                for ba in bin_arrs:
                    ref = ba.find(q0, namespaces=self.ns).get('ref')
                    q = 'm:referenceableParamGroup[@id="' + ref + '"]'
                    pg = pgr.find(q, namespaces=self.ns)
                    q = './/m:cvParam[@accession="MS:1000514"]/..'
                    if pg.find(q, namespaces=self.ns) is not None:
                        x = self.read_binary(ba, pgr)
                        break
                    q = './/m:cvParam[@accession="MS:1000617"]/..'
                    if pg.find(q, namespaces=self.ns) is not None:
                        x = self.read_binary(ba, pgr)
                        break
                    q = './/m:cvParam[@accession="MS:1000786"]/..'
                    if pg.find(q, namespaces=self.ns) is not None:
                        x = self.read_binary(ba, pgr)
                        break
                else:
                    continue

            q = './/m:cvParam[@accession="MS:1000515"]/..'
            y_elem = s.find(q, namespaces=self.ns)
            if y_elem is not None:
                y = self.read_binary(y_elem, pgr)
            else:
                # check paramGroupRef
                q0 = './/m:referenceableParamGroupRef'
                bin_arrs = s.findall(q0 + '/..', namespaces=self.ns)
                for ba in bin_arrs:
                    ref = ba.find(q0, namespaces=self.ns).get('ref')
                    q = 'm:referenceableParamGroup[@id="' + ref + '"]'
                    pg = pgr.find(q, namespaces=self.ns)
                    q = './/m:cvParam[@accession="MS:1000515"]/..'
                    if pg.find(q, namespaces=self.ns) is not None:
                        y = self.read_binary(ba, pgr)
                        break
                else:
                    continue

            yield Scan(x, y, name=time)

    def read_binary(self, ba, param_groups=None):
        """
        ba - binaryDataArray XML node
        """
        if ba is None:
            return []

        pgr = ba.find('m:referenceableParamGroupRef', namespaces=self.ns)
        if pgr is not None and param_groups is not None:
            q = 'm:referenceableParamGroup[@id="' + pgr.get('ref') + '"]'
            pg = param_groups.find(q, namespaces=self.ns)
        else:
            pg = ba

        if pg.find('m:cvParam[@accession="MS:1000574"]',
                   namespaces=self.ns) is not None:
            compress = True
        elif pg.find('m:cvParam[@accession="MS:1000576"]',
                     namespaces=self.ns) is not None:
            compress = False
        else:
            # TODO: no info? should check the other record?
            pass

        if pg.find('m:cvParam[@accession="MS:1000521"]',
                   namespaces=self.ns) is not None:
            dtype = 'f'
        elif pg.find('m:cvParam[@accession="MS:1000523"]',
                     namespaces=self.ns) is not None:
            dtype = 'd'
        else:
            # TODO: no info? should check the other record?
            pass

        datatext = ba.find('m:binary', namespaces=self.ns).text
        if compress:
            rawdata = zlib.decompress(base64.b64decode(datatext))
        else:
            rawdata = base64.b64decode(datatext)
        return np.fromstring(rawdata, dtype=dtype)

    def total_trace(self, twin=None):
        r = Et.parse(self.filename).getroot()

        # get it from the chromatogram list
        c = r.find('.//m:cvParam[@accession="MS:1000235"]/..',
                   namespaces=self.ns)
        if c is not None:
            q = './/m:cvParam[@accession="MS:1000595"]/..'
            index = self.read_binary(c.find(q, namespaces=self.ns))
            q = './/m:cvParam[@accession="MS:1000515"]/..'
            values = self.read_binary(c.find(q, namespaces=self.ns))
            return Trace(values, index, name='tic')

        # # otherwise try to extract it from the individual records
        # spectra = r.findall('*//m:spectrum/', namespaces=self.ns)
        # for s in spectra:
        #     # TIC
        #     s.find('m:cvParam[@accession="MS:1000285"]', namespaces=self.ns)
        #     # time
        #     s.find('.//m:cvParam[@accession="MS:1000016"]', \
        #            namespaces=self.ns)

        # TODO: call parent function if no TIC found


def write_mzxml(filename, df, info=None, precision='f'):
    """
    Precision is either f or d.
    """
    for r in df.values:
        df.columns
        pass


def write_mzml(filename, df, info=None):
    r = Et.Element('mzML')

    c = Et.SubElement(r, 'cvList', {'count': '1'})
    cva = {'id': 'MS', 'fullName': 'Proteomics Standards Initiative Mass ' +
           'Spectrometry Ontology', 'version': '1.18.2',
           'URI': 'http://psidev.cvs.sourceforge.net/*checkout*/psidev/psi' +
           '/psi-ms/mzML/controlledVocabulary/psi-ms.obo'}
    Et.SubElement(c, 'cv', cva)

    c = Et.SubElement(r, 'fileDescription')
    c = Et.SubElement(c, 'fileContent')
    acc = {'ms': ('MS:1000579', 'MS1 spectrum'),
           'uv': ('MS:1000806', 'absorption spectrum')}
    # TODO: need to select which kind of spectrum this is
    stype = 'ms'
    Et.SubElement(c, 'cvParam', {'cvRef': 'MS', 'accession': acc[stype][0],
                                 'name': acc[stype][1]})

    c = Et.SubElement(r, 'softwareList', {'count': '1'})
    c = Et.SubElement(c, 'software', {'id': 'Aston', 'version': __version__})
    Et.SubElement(c, 'cvParam', {'cvRef': 'MS', 'accession': 'MS:1001457',
                                 'name': 'data processing software'})

    c = Et.SubElement(r, 'instrumentConfigurationList', {'count': '1'})
    c = Et.SubElement(c, 'instrumentConfiguration', {'id': 'IC1'})
    Et.SubElement(c, 'cvParam', {'cvRef': 'MS', 'accession': 'MS:1000031',
                                 'name': 'instrument model'})

    c = Et.SubElement(r, 'dataProcessingList', {'count': '1'})
    c = Et.SubElement(c, 'dataProcessing', {'id': 'DP1'})
    c = Et.SubElement(c, 'processingMethod', {'order': '1',
                                              'softwareRef': 'Aston'})
    Et.SubElement(c, 'cvParam', {'cvRef': 'MS', 'accession': 'MS:1000544',
                                 'name': 'Conversion to mzML'})

    c = Et.SubElement(r, 'run', {'defaultInstrumentConfigurationRef': 'IC1',
                                 'id': 'R1'})
    sl = Et.SubElement(c, 'spectrumList', {'count': len(df.index),
                                           'defaultDataProcessingRef': 'DP1'})
    for i, scan in enumerate(df.scans()):
        s = Et.SubElement(sl, 'spectrum', {'index': str(i),
                                           'id': 'scan=' + str(i + 1),
                                           'defaultArrayLength': '1'})
        # TODO: write out spectrum info
        # ET.SubElement(s, 'scanList', {})  # is this necessary?
        dal = Et.SubElement(s, 'binaryDataArrayList', {'count': '2'})
        da1 = Et.SubElement(dal, 'binaryDataArray', {'encodedLength': ''})
        Et.SubElement(da1, 'cvParamGroup', {})
        Et.SubElement(da1, 'binary', {}).text = 'test'
        da2 = Et.SubElement(dal, 'binaryDataArray', {'encodedLength': ''})
        Et.SubElement(da2, 'cvParamGroup', {})
        Et.SubElement(da2, 'binary', {}).text = 'test'



from collections import OrderedDict
import numpy as np
from scipy.optimize import fsolve

# using 23.5 as a default d18o assuming no fractionation
# during combustion; actual combustion produces values
# lower than this. See Schumacher et al 2008 Atm Chem & Phys Dis


def delta13c_constants():
    """
    Constants for calculating delta13C values from ratios.
    From website of Verkouteren & Lee 2001 Anal. Chem.
    """
    # possible values for constants (from NIST)
    cst = OrderedDict()
    cst['Craig'] = {'S13': 0.0112372, 'S18': 0.002079,
                    'K': 0.008333, 'A': 0.5}
    cst['IAEA'] = {'S13': 0.0112372, 'S18': 0.00206716068,
                   'K': 0.0091993, 'A': 0.5}
    cst['Werner'] = {'S13': 0.0112372, 'S18': 0.0020052,
                     'K': 0.0093704, 'A': 0.516}
    cst['Santrock'] = {'S13': 0.0112372, 'S18': 0.0020052,
                       'K': 0.0099235, 'A': 0.516}
    cst['Assonov'] = {'S13': 0.0112372, 'S18': 0.0020052,
                      'K': 0.0102819162, 'A': 0.528}
    cst['Assonov2'] = {'S13': 0.0111802, 'S18': 0.0020052,
                       'K': 0.0102819162, 'A': 0.528}
    cst['Isodat'] = {'S13': 0.0111802, 'S18': 0.0020052,
                     'K': 0.0099235, 'A': 0.516}
    return cst


def delta13c_craig(r45sam, r46sam, d13cstd, r45std, r46std,
                   ks='Craig', d18ostd=23.5):
    """
    Algorithm from Craig 1957.

    From the original Craig paper, we can set up a pair of equations
    and solve for d13C and d18O simultaneously:

        d45 * r45 = r13 * d13
                  + 0.5 * r17 * d18
        d46 = r13 * ((r17**2 + r17 - r18) / a) * d13
            + 1 - 0.5 * r17 * ((r13**2 + r13 - r18) / a) * d18
        where a = r18 + r13 * r17 and b = 1 + r13 + r17
    """
    # the constants for the calculations
    # originally r13, r17, r18 = 1123.72e-5, 759.9e-6, 415.8e-5
    k = delta13c_constants()[ks]

    # TODO: not clear why need to multiply by 2?
    r13, r18 = k['S13'], 2 * k['S18']
    r17 = 2 * (k['K'] * k['S18'] ** k['A'])
    a = (r18 + r13 * r17) * (1. + r13 + r17)

    # the coefficients for the calculations
    eqn_mat = np.array([[r13, 0.5 * r17],
                        [r13 * ((r17 ** 2 + r17 - r18) / a),
                         1 - 0.5 * r17 * ((r13 ** 2 + r13 - r18) / a)]])

    # precalculate the d45 and d46 of the standard versus PDB
    r45d45std = (eqn_mat[0, 0] * d13cstd + eqn_mat[0, 1] * d18ostd)
    d46std = eqn_mat[1, 0] * d13cstd + eqn_mat[1, 1] * d18ostd

    # calculate the d45 and d46 of our sample versus PDB
    # in r45d45, r45 of PDB = r13 + r17 of PDB
    r45d45 = 1000. * (r45sam / r45std - 1.) * \
        (r13 + r17 + 0.001 * r45d45std) + r45d45std
    d46 = 1000. * (r46sam / r46std - 1.) * (1. + 0.001 * d46std) + d46std

    # solve the system of equations
    x = np.linalg.solve(eqn_mat, np.array([r45d45, d46]))
    return x[0]


def delta13c_santrock(r45sam, r46sam, d13cstd, r45std, r46std,
                      ks='Santrock', d18ostd=23.5):
    """
    Given the measured isotope signals of a sample and a
    standard and the delta-13C of that standard, calculate
    the delta-13C of the sample.

    Algorithm from Santrock, Studley & Hayes 1985 Anal. Chem.
    """
    k = delta13c_constants()[ks]

    # function for calculating 17R from 18R
    def c17(r):
        return k['K'] * r ** k['A']
    rcpdb, rosmow = k['S13'], k['S18']

    # known delta values for the ref peak
    r13std = (d13cstd / 1000. + 1) * rcpdb
    r18std = (d18ostd / 1000. + 1) * rosmow

    # determine the correction factors
    c45 = r13std + 2 * c17(r18std)
    c46 = c17(r18std) ** 2 + 2 * r13std * c17(r18std) + 2 * r18std

    # correct the voltage ratios to ion ratios
    r45 = (r45sam / r45std) * c45
    r46 = (r46sam / r46std) * c46

    def rf(r18):
        return -3 * c17(r18) ** 2 + 2 * r45 * c17(r18) + 2 * r18 - r46
    # r18 = scipy.optimize.root(rf, r18std).x[0]  # use with scipy 0.11.0
    r18 = fsolve(rf, r18std)[0]
    r13 = r45 - 2 * c17(r18)
    return 1000 * (r13 / rcpdb - 1)


def delta13c_brand(r45sam, r46sam, d13cstd, r45std, r46std):
    """
    Linear simplification from Brand et al. 2010 Pure Appl Chem
    """
    # d13C = d45 + 2 * 0.035168 * (d45 - d46)
    raise NotImplementedError



# -*- coding: utf-8 -*-

import os.path as op
import gzip
import io
import struct
from datetime import datetime
from xml.etree import ElementTree
import numpy as np
import scipy.sparse
from aston.cache import cache
from aston.trace import Trace, Chromatogram
from aston.tracefile import TraceFile, ScanListFile
from aston.spectra import Scan


class AgilentMS(TraceFile):
    mime = 'application/vnd-agilent-chemstation-ms'
    traces = ['#ms']

    def total_trace(self, twin=None):
        # TODO: use twin?
        f = open(self.filename, 'rb')

        # get number of scans to read in
        f.seek(0x5)
        if f.read(4) == 'GC':
            f.seek(0x142)
        else:
            f.seek(0x118)
        nscans = struct.unpack('>H', f.read(2))[0]

        # find the starting location of the data
        f.seek(0x10A)
        f.seek(2 * struct.unpack('>H', f.read(2))[0] - 2)

        tme = np.zeros(nscans)
        tic = np.zeros(nscans)
        for i in range(nscans):
            npos = f.tell() + 2 * struct.unpack('>H', f.read(2))[0]
            tme[i] = struct.unpack('>I', f.read(4))[0] / 60000.
            f.seek(npos - 4)
            tic[i] = struct.unpack('>I', f.read(4))[0]
            f.seek(npos)
        f.close()
        return Trace(tic, tme, name='TIC')

    @property
    @cache(maxsize=1)
    def data(self):
        f = open(self.filename, 'rb')

        # get number of scans to read in
        # note that GC and LC chemstation store this in slightly different
        # places
        f.seek(0x5)
        if f.read(4) == 'GC':
            f.seek(0x142)
        else:
            f.seek(0x118)
        nscans = struct.unpack('>H', f.read(2))[0]

        f.seek(0x10A)
        f.seek(2 * struct.unpack('>H', f.read(2))[0] - 2)
        dstart = f.tell()

        # determine total number of measurements in file
        tot_pts = 0
        rowst = np.empty(nscans + 1, dtype=int)
        rowst[0] = 0

        for scn in range(nscans):
            # get the position of the next scan
            npos = f.tell() + 2 * struct.unpack('>H', f.read(2))[0]

            # keep a running total of how many measurements
            tot_pts += (npos - f.tell() - 26) // 4
            rowst[scn + 1] = tot_pts

            # move forward
            f.seek(npos)

        # go back to the beginning and load all the other data
        f.seek(dstart)

        ions = []
        i_lkup = {}
        cols = np.empty(tot_pts, dtype=int)
        vals = np.empty(tot_pts, dtype=np.int32)
        times = np.empty(nscans)

        for scn in range(nscans):
            npos = f.tell() + 2 * struct.unpack('>H', f.read(2))[0]

            # the sampling rate is evidentally 60 kHz on all Agilent's MS's
            times[scn] = struct.unpack('>I', f.read(4))[0] / 60000.

            f.seek(f.tell() + 12)
            npts = rowst[scn + 1] - rowst[scn]
            mzs = struct.unpack('>' + npts * 'HH', f.read(npts * 4))

            # there's some bug in the numpy implementation that makes this fail
            # after the first time
            # mzs = np.fromfile(f, dtype='>H', count=npts * 2)

            nions = set(mzs[0::2]).difference(i_lkup)
            i_lkup.update({ion: i + len(ions) for i, ion in enumerate(nions)})
            ions += nions

            cols[rowst[scn]:rowst[scn + 1]] = [i_lkup[i] for i in mzs[0::2]]
            vals[rowst[scn]:rowst[scn + 1]] = mzs[1::2]
            f.seek(npos)
        f.close()

        vals = ((vals & 16383) * 8 ** (vals >> 14)).astype(float)
        data = scipy.sparse.csr_matrix((vals, cols, rowst),
                                       shape=(nscans, len(ions)), dtype=float)
        ions = np.array(ions) / 20.
        return Chromatogram(data, times, ions)

    @property
    @cache(maxsize=1)
    def old_data(self):
        f = open(self.filename, 'rb')

        # get number of scans to read in
        # note that GC and LC chemstation store this in slightly different
        # places
        f.seek(0x5)
        if f.read(4) == 'GC':
            f.seek(0x142)
        else:
            f.seek(0x118)
        nscans = struct.unpack('>H', f.read(2))[0]

        # find the starting location of the data
        f.seek(0x10A)
        f.seek(2 * struct.unpack('>H', f.read(2))[0] - 2)

        # make a list of all of the ions and also read in times
        ions = set()
        times = np.empty(nscans)
        scan_locs = np.empty(nscans, dtype=int)
        scan_pts = np.empty(nscans, dtype=int)
        for scn in range(nscans):
            npos = f.tell() + 2 * struct.unpack('>H', f.read(2))[0]

            # the sampling rate is evidentally 60 kHz on all Agilent's MS's
            times[scn] = struct.unpack('>I', f.read(4))[0] / 60000.

            f.seek(f.tell() + 6)
            npts = struct.unpack('>H', f.read(2))[0]

            # jump to the data and save relevant parameters for later
            f.seek(f.tell() + 4)
            scan_locs[scn] = f.tell()
            scan_pts[scn] = npts

            # TODO: use numpy.fromfile?
            nions = np.fromfile(f, dtype='>H', count=npts * 2)[0::2]
            # if scn < 2:
            #     print(npts)
            #     print(nions)
            # nions = struct.unpack('>' + npts * 'HH', f.read(npts * 4))[0::2]
            ions.update(nions)
            f.seek(npos)

        ions = np.array(sorted(list(ions)))
        data = np.empty((len(times), len(ions)), dtype=float)
        for scn in range(nscans):
            f.seek(scan_locs[scn])
            # TODO: use numpy.fromfile?
            mzs = np.fromfile(f, dtype='>H', count=scan_pts[scn] * 2)
            # mzs = np.array(struct.unpack('>' + npts * 'HH', f.read(npts * 4)))  # noqa
            if len(mzs) == 0:
                continue
            ilocs = np.searchsorted(ions, mzs[0::2])
            abn = (mzs[1::2] & 16383) * 8 ** (mzs[1::2] >> 14)
            data[scn][ilocs] = abn
        f.close()

        ions /= 20.
        return Chromatogram(data, times, ions)

    @property
    def info(self):
        d = super(AgilentMS, self).info
        f = open(self.filename, 'rb')
        f.seek(0x18)
        d['name'] = f.read(struct.unpack('>B', f.read(1))[0]).decode().strip()
        f.seek(0x94)
        d['r-opr'] = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        f.seek(0xE4)
        d['m-name'] = f.read(struct.unpack('>B', f.read(1))[0]).decode().strip()  # noqa
        f.seek(0xB2)
        rawdate = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        try:
            d['r-date'] = datetime.strptime(rawdate,
                                            "%d %b %y %H:%M %p").isoformat(' ')
        except ValueError:
            pass  # date is not in correct format to parse?
        # TODO: vial number in here too?
        f.close()

        # TODO: fill this out
        # read info from the acqmeth.txt file
        # fname = op.join(op.dirname(self.filename), 'acqmeth.txt')

        return d


class AgilentMSMSScan(ScanListFile):
    mime = 'application/vnd-agilent-masshunter-msmsscan'
    traces = ['#ms']

    # TODO: __init__ method that adds mrm trace names to traces
    def _scan_iter(self, keylist):
        f = open(self.filename, 'rb')
        r = ElementTree.parse(op.splitext(self.filename)[0] + '.xsd').getroot()

        xml_to_struct = {'xs:int': 'i', 'xs:long': 'q', 'xs:short': 'h',
                         'xs:byte': 'b', 'xs:double': 'd', 'xs:float': 'f'}
        rfrmt = {}

        for n in r.getchildren():
            name = n.get('name')
            for sn in n.getchildren()[0].getchildren():
                if rfrmt.get(name, None) is None:
                    rfrmt[name] = []
                sname = sn.get('name')
                stype = sn.get('type')
                rfrmt[name].append((sname, xml_to_struct.get(stype, stype)))

        def resolve(lookup, recname):
            names = [i[0] for i in lookup[recname]]
            frmts = [i[1] for i in lookup[recname]]
            flatnames = []
            flatfrmts = ''
            for n, f in zip(names, frmts):
                if len(f) != 1:
                    n, f = resolve(lookup, f)
                    flatnames += n
                else:
                    flatnames.append(n)
                flatfrmts += f
            return flatnames, flatfrmts

        fnames, ffrmts = resolve(rfrmt, 'ScanRecordType')
        rec_str = '<' + ffrmts
        sz = struct.calcsize(rec_str)

        f.seek(0x58)
        start_offset = struct.unpack('<i', f.read(4))[0]
        f.seek(start_offset)

        locations = [fnames.index(k) for k in keylist]
        while True:
            try:
                data = struct.unpack(rec_str, f.read(sz))
            except struct.error:
                break
            yield (data[loc] for loc in locations)
        f.close()

    def total_trace(self, twin=None):
        if twin is None:
            twin = (-np.inf, np.inf)
        tme = []
        tic = []
        for t, z in self._scan_iter(['ScanTime', 'TIC']):
            if t < twin[0]:
                continue
            elif t > twin[1]:
                break
            tme.append(t)
            tic.append(z)
        return Trace(np.array(tic), np.array(tme), name='TIC')
        # TODO: set .twin(twin) bounds on this

    def scans(self, twin=None):
        if twin is None:
            twin = (-np.inf, np.inf)

        # super hack-y way to disable checksum and length checking
        gzip.GzipFile._read_eof = lambda _: None  # noqa
        # standard prefix for every zip chunk
        gzprefix = b'\x1f\x8b\x08\x00\x00\x00\x00\x00\x04\x00'

        def uncompress(d):
            return gzip.GzipFile(fileobj=io.BytesIO(gzprefix + d)).read()

        f = open(op.join(op.split(self.filename)[0], 'MSProfile.bin'), 'rb')

        flds = ['ScanTime', 'SpectrumFormatID', 'SpectrumOffset',
                'ByteCount', 'PointCount', 'MinX', 'MaxX']

        for t, fmt, off, bc, pc, minx, maxx in self._scan_iter(flds):
            if t < twin[0]:
                continue
            if t > twin[1]:
                break

            f.seek(off)
            if fmt == 1:
                # this record is compressed with gz
                profdata = uncompress(f.read(bc))
                pd = np.array(struct.unpack('dd' + pc * 'i', profdata)[2:])
            elif fmt == 2:
                profdata = f.read(bc)
                pd = np.array(struct.unpack('dd' + pc * 'f', profdata)[2:])
            else:
                raise NotImplementedError('Unknown Agilent MH Scan format')
            # TODO: probably not a good approximation?
            ions = np.linspace(minx, maxx, len(pd))
            yield Scan(ions, pd, name=t)

        f.close()

    def mrm_trace(self, parent=None, daughter=None, tol=0.5, twin=None):
        # TODO: should override `trace` and then call parent's `trace` method
        # if name is not an mrm trace
        if twin is None:
            twin = (-np.inf, np.inf)

        tme, ic = [], []
        for t, off, bc, pc, minx, maxx, d_mz, p_mz, z in self._scan_iter([
            'ScanTime', 'SpectrumOffset', 'ByteCount', 'PointCount',
            'MinX', 'MaxX', 'BasePeakMZ', 'MzOfInterest', 'TIC'
        ]):
            if t < twin[0]:
                continue
            elif t > twin[1]:
                break
            if parent is not None:
                if np.abs(parent - p_mz) > tol:
                    continue
            if daughter is not None:
                if np.abs(daughter - d_mz) > tol:
                    continue
            tme.append(t)
            ic.append(z)

        return Trace(np.array(ic), np.array(tme),
                     name=str(str(parent) + '→' + str(daughter)))



# -*- coding: utf-8 -*-
import struct
import numpy as np
from aston.cache import cache
from aston.trace import Trace
from aston.tracefile import TraceFile


class AgilentCSPump(TraceFile):
    mime = 'application/vnd-agilent-chemstation-pump'
    traces = []
    # traces = ['mslva', 'mslvb', 'mslvc', 'mslvd', 'mflow']

    @property
    def info(self):
        d = super(AgilentCSPump, self).info
        fd = read_reg_file(open(self.filename, 'rb'))
        if fd.get('TIMETABLE', False):
            # get solv_B, solv_C, solv_D, flow
            # calculate solv_A
            pass
        else:
            if fd.get('SOLV_RATIO_A', False):
                pass
            if fd.get('SOLV_RATIO_B', False):
                pass
            if fd.get('SOLV_RATIO_C', False):
                pass
            if fd.get('SOLV_RATIO_D', False):
                pass
            if fd.get('FLOW', False):
                pass
        return d

    def _trace(self, name, twin):
        # TODO: return method traces
        pass


class AgilentCSFraction(TraceFile):
    mime = 'application/vnd-agilent-chemstation-fraction'
    traces = ['*fxn']

    def events(self, name='fxn', twin=None):
        # TODO: use twin
        evts = super(AgilentCSFraction, self).events(name, twin)
        if name != 'fxn':
            return evts

        acq_file = open(self.filename, 'rb')
        acq_file.seek(0x19)
        vers = acq_file.read(7)

        if vers.startswith(b'A'):
            print('FXN', self.filename)
        elif vers == b'notused':
            acq_file.seek(0xED)
            nfxns = struct.unpack('<I', acq_file.read(4))[0]
            acq_file.seek(0x33F)
            for _ in range(nfxns):
                d = struct.unpack('<Q9fI4f7I', acq_file.read(92))
                evts.append({'t0': d[13] / 60000., 't1': d[14] / 60000.})

            acq_file.seek(acq_file.tell() + 264)
            for i in range(nfxns):
                slen = struct.unpack('<H', acq_file.read(2))[0] - 1
                evts[i]['name'] = acq_file.read(2 * slen).decode('utf-16')
        return evts


class AgilentCSFlowInject(TraceFile):
    mime = 'application/vnd-agilent-chemstation-flowinject'
    traces = ['*fia']

    def events(self, name='fia', twin=None):
        # TODO: use twin
        evts = super(AgilentCSFlowInject, self).events(name, twin)
        if name != 'fia':
            return evts

        acq_file = open(self.filename, 'rb')
        acq_file.seek(0x19)
        vers = acq_file.read(7)

        if vers.startswith(b'A'):
            d = read_reg_file(acq_file)
            if not d.get(u'FIARun', False):
                return []
            prev_f = d['FIASeriesInfo'][1][1]
            for f in d['FIASeriesInfo'][1][2:]:
                evts.append({'t0': prev_f[0], 't1': f[0], 'name': prev_f[2]})
                prev_f = f
            else:
                if len(evts) > 0:
                    off_t = evts[-1]['t1'] - evts[-1]['t0']
                    evts.append({'t0': prev_f[0], 't1': prev_f[0] + off_t,
                                 'name': prev_f[2]})
        elif vers == b'notused':
            # TODO: get fia from new-style *.REG files.
            print('FIA', self.filename)
            pass
        return evts


class AgilentCSLC(TraceFile):
    mime = 'application/vnd-agilent-chemstation-lcstat'

    _tr_names = {'pres': 'PMP1, Pressure', 'flow': 'PMP1, Flow',
                 'slva': 'PMP1, Solvent A', 'slvb': 'PMP1, Solvent B',
                 'slvc': 'PMP1, Solvent C', 'slvd': 'PMP1, Solvent D'}

    @property
    @cache(maxsize=1)
    def traces(self):
        traces = []
        for abrv, full in self._tr_names.items():
            df = read_multireg_file(open(self.filename, 'rb'), title=full)
            if 'Trace' in df:
                traces.append(abrv)
        return traces

    def _trace(self, name, twin):
        # TODO: read info from new style REG files
        df = read_multireg_file(open(self.filename, 'rb'),
                                title=self._tr_names[name])
        if 'Trace' not in df:
            return Trace([], [])
        ts = df['Trace']
        ts.name = name
        return ts.twin(twin)


def read_multireg_file(f, title=None):
    """
    Some REG files have multiple "sections" with different data.
    This parses each chunk out of such a file (e.g. LCDIAG.REG)
    """
    f.seek(0x26)
    nparts = struct.unpack('<H', f.read(2))[0]
    foff = 0x2D
    if title is None:
        data = []
        for _ in range(nparts):
            d = read_reg_file(f, foff)
            data.append(d)
            foff = f.tell() + 1
    else:
        for _ in range(nparts):
            d = read_reg_file(f, foff)
            if d.get('Title') == title:
                data = d
                break
            foff = f.tell() + 1
        else:
            data = {}
    return data


def read_reg_file(f, foff=0x2D):
    """
    Given a file handle for an old-style Agilent *.REG file, this
    will parse that file into a dictonary of key/value pairs
    (including any tables that are in the *.REG file, which will
    be parsed into lists of lists).
    """
    # convenience function for reading in data
    def rd(st):
        return struct.unpack(st, f.read(struct.calcsize(st)))

    f.seek(0x19)
    if f.read(1) != b'A':
        # raise TypeError("Version of REG file is too new.")
        return {}

    f.seek(foff)
    nrecs = rd('<I')[0]  # TODO: should be '<H'
    rec_tab = [rd('<HHIII') for n in range(nrecs)]

    names = {}
    f.seek(foff + 20 * nrecs + 4)
    for r in rec_tab:
        d = f.read(r[2])
        if r[1] == 1539:  # '0306'
            # this is part of the linked list too, but contains a
            # reference to a table
            cd = struct.unpack('<HIII21sI', d)
            names[cd[5]] = cd[4].decode('iso8859').strip('\x00')
            # except:
            #     pass
        elif r[1] == 32769 or r[1] == 32771:  # b'0180' or b'0380'
            names[r[4]] = d[:-1].decode('iso8859')
        elif r[1] == 32774:  # b'0680'
            # this is a string that is referenced elsewhere (in a table)
            names[r[4]] = d[2:-1].decode('iso8859')
        elif r[1] == 32770:  # b'0280'
            # this is just a flattened numeric array
            names[r[4]] = np.frombuffer(d, dtype=np.uint32, offset=4)

    data = {}
    f.seek(foff + 20 * nrecs + 4)
    for r in rec_tab:
        d = f.read(r[2])
        if r[1] == 1538:  # '0206'
            # this is part of a linked list
            if len(d) == 43:
                cd = struct.unpack('<HIII21sd', d)
                data[cd[4].decode('iso8859').strip('\x00')] = cd[5]
            else:
                pass
        elif r[1] == 1537:  # b'0106'
            # name of property
            n = d[14:30].split(b'\x00')[0].decode('iso8859')
            # with value from names
            data[n] = names.get(struct.unpack('<I', d[35:39])[0], '')
        elif r[1] == 1793:  # b'0107'
            # this is a table of values
            nrow = struct.unpack('<H', d[4:6])[0]
            ncol = struct.unpack('<H', d[16:18])[0]
            if ncol != 0:
                cols = [struct.unpack('<16sHHHHHI', d[20 + 30 * i:50 + 30 * i])
                        for i in range(ncol)]
                colnames = [c[0].split(b'\x00')[0].decode('iso8859')
                            for c in cols]
                # TODO: type 2 is not a constant size? 31, 17
                rty2sty = {1: 'H', 3: 'I', 4: 'f', 5: 'H',
                           7: 'H', 8: 'd', 11: 'H', 12: 'H',
                           13: 'I', 14: 'I', 16: 'H'}
                coltype = '<' + ''.join([rty2sty.get(c[3], str(c[2]) + 's')
                                         for c in cols])
                lencol = struct.calcsize(coltype)
                tab = []
                for i in reversed(range(2, nrow + 2)):
                    rawrow = struct.unpack(coltype,
                                           d[-i * lencol: (1 - i) * lencol])
                    row = []
                    for j, p in enumerate(rawrow):
                        if cols[j][3] == 3:
                            row.append(names.get(p, str(p)))
                        else:
                            row.append(p)
                    tab.append(row)
                data[names[r[4]]] = [colnames, tab]
        elif r[1] == 1281 or r[1] == 1283:  # b'0105' or b'0305'
            fm = '<HHBIIhIdII12shIddQQB8sII12shIddQQB8s'
            m = struct.unpack(fm, d)
            nrecs = m[4]  # number of points in table

            # x_units = names.get(m[8], '')
            x_arr = m[14] * names.get(m[9], np.arange(nrecs - 1))
            y_arr = m[25] * names.get(m[20])
            y_units = names.get(m[19], '')
            if y_units == 'bar':
                y_arr *= 0.1  # convert to MPa
            # TODO: what to call this?
            data['Trace'] = Trace(y_arr, x_arr, name='')
        # elif r[1] == 1025:  # b'0104'
        #     # lots of zeros? maybe one or two numbers?
        #     # only found in REG entries that have long 0280 records
        #     fm = '<HQQQIHHHHIIHB'
        #     m = struct.unpack(fm, d)
        #     print(m)
        #     #print(r[1], len(d), binascii.hexlify(d))
        #     pass
        # elif r[1] == 512:  # b'0002'
        #     # either points to two null pointers or two other pointers
        #     # (indicates start of linked list?)
        #     print(r[1], len(d), binascii.hexlify(d))
        # elif r[1] == 769 or r[1] == 772:  # b'0103' or b'0403'
        #     # points to 2nd, 3rd & 4th records (two 0002 records and a 0180)
        #     b = binascii.hexlify
        #     print(b(d[10:14]), b(d[14:18]), b(d[18:22]))

    return data


def read_regb_file(f):
    pass
    # CObArray
    # either 2 bytes long (0100) or 8 (0000 0380 0580 XXXX)

    # CHPAnnText
    # [j[36:36+2*struct.unpack('<H', j[34:36])[0]].decode('utf-16') \
    #  for i, j in a if i == b'CHPAnnText']
    # slen = struct.unpack('<H', struct_dat[34:36])
    # struct_dat[36:36 + slen * 2].decode('utf-16')

    # CHPDatLongRow
    #  starts at byte 14?

    # CHPNdrString
    # [j[12:12 + 2 * struct.unpack('<H', j[10:12])[0]].decode('utf-16') \
    #  for i, j in a if i == b'CHPNdrString']

    # CHPNdrObject

    # CHPTable



# -*- coding: utf-8 -*-
import struct
from xml.etree import ElementTree
import numpy as np
from aston.trace import Trace
from aston.tracefile import TraceFile


class AgilentMHPump(TraceFile):
    mime = 'application/vnd-agilent-masshunter-pump'
    traces = ['pres', 'flow', 'slvb']

    def _trace(self, name, twin):
        return read_mh_trace(self.filename, name).twin(twin)


class AgilentMHTemp(TraceFile):
    mime = 'application/vnd-agilent-masshunter-temp'
    traces = ['temp']

    def _trace(self, name, twin):
        return read_mh_trace(self.filename, name).twin(twin)


class AgilentMHAcqMethod(TraceFile):
    mime = 'application/vnd-agilent-masshunter-acqmethod'

    @property
    def info(self):
        d = super(AgilentMHAcqMethod, self).info
        try:
            r = ElementTree.parse(self.filename).getroot()
            d['r-en-time'] = r.find('.//CapPump//Stoptime').text
            d['m-flw'] = r.find('.//CapPump//Flow').text
            d['m-slv'] = r.find('.//CapPump//SolvNameA').text
            d['m-slv-b'] = r.find('.//CapPump//SolvNameB').text
            d['m-slv-b-per'] = r.find('.//CapPump//SolvRatioB').text
            d['m-slv-c'] = r.find('.//CapPump//SolvNameC').text
            d['m-slv-d'] = r.find('.//CapPump//SolvNameD').text
            d['m-tmp'] = r.find('.//TCC//LeftTemp').text
        except AttributeError:
            # e.g. if LeftTemp is not set, find will
            # return None and None has no attribute text
            # TODO: better fix for this
            pass
        return d


class AgilentMHSampleInfo(TraceFile):
    mime = 'application/vnd-agilent-masshunter-sampleinfo'

    @property
    def info(self):
        try:
            def u(s):
                return s.decode('utf-8')
            u('')
        except AttributeError:
            def u(s):
                return s

        d = super(AgilentMHSampleInfo, self).info
        r = ElementTree.parse(self.filename).getroot()
        info = {i.find('Name').text: i.find('Value').text
                for i in r.findall('Field')}
        d['r-smp'] = info.get('Sample Name', '')
        d['r-vial-pos'] = info.get('Sample Position', '')
        d['r-inst'] = info.get('InstrumentName', '')
        d['r-opr'] = info.get('OperatorName', '')
        d['r-date'] = info.get('AcqTime', '').replace('T', ' ').rstrip('Z')
        d['m-inj-size'] = info.get(u('Inj Vol (µl)'), '')
        return d


def read_mh_trace(filename, trace_name):
    f = open(filename, 'rb')
    fdat = open(filename[:-3] + '.cg', 'rb')

    ttab = {'pres': 'Pressure', 'flow': 'Flow', 'slvb': '%B',
            'temp': 'Temperature of Left Heat Exchanger'}

    # convenience function for reading in data
    def rd(st):
        return struct.unpack(st, f.read(struct.calcsize(st)))

    f.seek(0x4c)
    num_traces = rd('<I')[0]
    for _ in range(num_traces):
        cloc = f.tell()
        f.seek(cloc + 2)
        sl = rd('<B')[0]
        cur_trace_name = rd('<' + str(sl) + 's')[0]
        if ttab[trace_name] == cur_trace_name:
            f.seek(f.tell() + 4)
            foff = rd('<Q')[0]
            npts = rd('<I')[0] + 2  # +2 for the extra time info
            fdat.seek(foff)
            pts = struct.unpack('<' + npts * 'd', fdat.read(8 * npts))
            # TODO: pts[0] is not the true offset?
            t = pts[0] + pts[1] * np.arange(npts - 2)
            d = np.array(pts[2:])
            # get the units
            f.seek(f.tell() + 40)
            sl = rd('<B')[0]
            y_units = rd('<' + str(sl) + 's')[0]
            if y_units == 'bar':
                d *= 0.1  # convert to MPa for metricness
            elif y_units == '':
                pass  # TODO: ul/min to ml/min
            return Trace(d, t, name=trace_name)

        f.seek(cloc + 87)



import struct
import numpy as np
import scipy.sparse
from aston.trace import Chromatogram, Trace
from aston.tracefile import TraceFile


class BrukerMSMS(TraceFile):
    mime = 'application/vnd-bruker-msms'
    traces = ['#ms']

    # def _getTotalTrace(self):
    #     pass

    @property
    def data(self):
        # convenience function for reading in data
        def rd(f, st):
            return struct.unpack(st, f.read(struct.calcsize(st)))

        # open the file
        f = open(self.filename, 'rb')

        nscans = rd(f, 'ii')[1]
        if nscans == 0:
            self.data = Trace(np.array([]), np.array([]), [])
            return
        times = np.array(rd(f, nscans * 'd')) / 60.0
        f.seek(f.tell() + 4)  # number of scans again

        # set up the array of column indices
        indptr = np.empty(nscans + 1, dtype=int)
        indptr[0] = 0

        # figure out the total number of points
        dpos = f.tell()
        tot_pts = 0
        for scn in range(nscans):
            npts = rd(f, 'i')[0]
            # rd(f, npts * 'f' + 'i' + n_pts * 'f')
            f.seek(f.tell() + 8 * npts + 4)
            tot_pts += npts
            indptr[scn + 1] = tot_pts
        f.seek(dpos)

        ions = []
        i_lkup = {}
        idxs = np.empty(tot_pts, dtype=int)
        vals = np.empty(tot_pts, dtype=float)

        for scn in range(nscans):
            npts = rd(f, 'i')[0]
            rd_ions = rd(f, npts * 'f')
            f.seek(f.tell() + 4)  # number of points again
            abun = rd(f, npts * 'f')

            nions = set([int(i) for i in rd_ions if int(i) not in i_lkup])
            i_lkup.update(dict((ion, i + len(ions) - 1)
                               for i, ion in enumerate(nions)))
            ions += nions

            idxs[indptr[scn]:indptr[scn + 1]] = \
                [i_lkup[int(i)] for i in rd_ions]
            vals[indptr[scn]:indptr[scn + 1]] = \
                abun

        idxs += 1
        data = scipy.sparse.csr_matrix((vals, idxs, indptr),
                                       shape=(nscans, len(ions)), dtype=float)
        return Chromatogram(data, times, ions)

        # self.data = np.zeros((recs, 2))
        # times = rd(f, nscans * 'd')
        # self.data[:, 0] = np.array(times) / 60
        # ions = set()
        # rd(f, 'i')  # number of data points again
        # for i in range(nscans):
        #     n_pts = rd(f, 'i')[0]
        #     ions.update(rd(f, n_pts * 'f'))
        #     rd(f, 'i')  # number of pts in spectra again
        #     abun = rd(f, n_pts * 'f')
        #     #self.data[i, 0] = times[i] / 60
        #     self.data[i, 1] = sum(abun)
        # f.close()
        # self.ions = [1]


class BrukerBAF(TraceFile):
    mime = 'application/vnd-bruker-baf'

    pass
    # TODO: implement this
    # 0x000c - q - 230 or 242

    ###############################################
    # file 1 - 230 ("Carolina")
    # 0xFFFFFFFFFFFF at 0x678D, 0x6825, 0xC459, 0x491AD,
    # 0x500E7, 0x57C39, and 25+ others

    # text section starts 0x018E, ends 0x6708

    # 0x409BF - (d- -1.000) then a ton of doubles

    # 3000 scans?, 2371 ions

    ###############################################
    # file 2 - 230 ("Short")

    ###############################################
    # file 3 - 242
    # 0xFFFFFFFFFFFF at 0x6BEF, 0x6CB7, 0x4044A

    # text section starts 0x0186, ends 0x6B3B

    # some CHP records at 0xAE9B til 0xBF5d (?)

    # 0x42a1e - (d- -1.000) then a ton of doubles
    # til 0x23DFFF56

    # 0x23FA916A - 0xFFFFFFFF before last data chunk?



import numpy as np
from scipy.sparse import dia_matrix
from scipy.sparse import lil_matrix

# TODO: implementations of Stein & Scott 1994 algorithms


def find_spectrum_match(spec, spec_lib, method='euclidian'):
    """
    Find spectrum in spec_lib most similar to spec.
    """
    # filter out any points with abundance below 1 %
    # spec[spec / np.sum(spec) < 0.01] = 0
    # normalize everything to sum to 1
    spec = spec / np.max(spec)

    if method == 'dot':
        d1 = (spec_lib * lil_matrix(spec).T).sum(axis=1).A ** 2
        d2 = np.sum(spec ** 2) * spec_lib.multiply(spec_lib).sum(axis=1).A
        dist = d1 / d2
    elif method == 'euclidian':
        # st_spc = spectrum[np.newaxis, :].repeat(spec_lib.shape[0], axis=0)
        st_spc = dia_matrix((spec, [0]), shape=(len(spec), len(spec)))
        # calculate the residual sum of squares from spectrum to library
        dist_sp = spec_lib.multiply(spec_lib) - 2 * spec_lib.dot(st_spc)
        dist = dist_sp.sum(axis=1).A + np.sum(spec ** 2)
    return (dist.argmin(), dist.min())



import numpy as np
from aston.trace import Trace
from aston.trace.math_traces import movingaverage


def molmz(df, noise=10000):
    """
    The mz of the molecular ion.
    """
    d = ((df.values > noise) * df.columns).max(axis=1)
    return Trace(d, df.index, name='molmz')


def mzminus(df, minus=0, noise=10000):
    """
    The abundances of ions which are minus below the molecular ion.
    """
    mol_ions = ((df.values > noise) * df.columns).max(axis=1) - minus
    mol_ions[np.abs(mol_ions) < 0] = 0
    d = np.abs(np.ones(df.shape) * df.columns -
               (mol_ions[np.newaxis].T * np.ones(df.shape))) < 1
    d = (df.values * d).sum(axis=1)
    return Trace(d, df.index, name='m-' + str(minus))


def basemz(df):
    """
    The mz of the most abundant ion.
    """
    # returns the
    d = np.array(df.columns)[df.values.argmax(axis=1)]
    return Trace(d, df.index, name='basemz')


def coda(df, window, level):
    """
    CODA processing from Windig, Phalp, & Payne 1996 Anal Chem
    """
    # pull out the data
    d = df.values

    # smooth the data and standardize it
    smooth_data = movingaverage(d, df.index, window)[0]
    stand_data = (smooth_data - smooth_data.mean()) / smooth_data.std()

    # scale the data to have unit length
    scale_data = d / np.sqrt(np.sum(d ** 2, axis=0))

    # calculate the "mass chromatographic quality" (MCQ) index
    mcq = np.sum(stand_data * scale_data, axis=0) / np.sqrt(d.shape[0] - 1)

    # filter out ions with an mcq below level
    good_ions = [i for i, q in zip(df.columns, mcq) if q >= level]
    return good_ions



# -*- coding: utf-8 -*-
"""
Functions which model peaks and decorator to allow those functions
to be translated and scaled and have the bounds of their parameters
respected.

Unless otherwise noted, the formulation of how each function is
written is taken from DiMarco & Bombi 2001 J Chrom A

Example:
    t = np.linspace(-10,10,200)
    peak = gaussian(t, x=2, w=2)
    fitted_peak = lorentzian(t, **fit_to(lorentzian, t, l))
    plt.plot(t, peak, 'k-', t, fitted_peak, 'r-')
    plt.show()
"""
import inspect
from functools import wraps
import numpy as np
from numpy import exp, sqrt, abs, log
from scipy.special import erfc, i1, gamma


# These functions allow us to use commonsense notation
# in labelling parameter bounds. The sqrt is to allow
# the bounded parameter to be multiplied without fear of it
# rounding outside the bounds.
def openhi(i):
    return np.nextafter(i, -1)


def openlow(i):
    return i + np.sqrt(np.nextafter(i, 1) - i)


def bounds(**kwargs):
    """
    Using this function as a decorator allow us to label which
    parameters of a function shouldn't have their values allowed
    to fall outside of a specific range.
    """
    def wrap(f):
        f._pbounds = kwargs
        return f
    return wrap


def peak_model(f):
    """
    Given a function that models a peak, add scale and location arguments to

    For all functions, v is vertical offset, h is height
    x is horizontal offset (1st moment), w is width (2nd moment),
    s is skewness (3rd moment), e is excess (4th moment)
    """
    @wraps(f)
    def wrapped_f(t, **kw):
        # load kwargs with default values
        # do this here instead of in the def because we want to parse
        # all of kwargs later to copy values to pass into f
        def_vals = {'v': 0.0, 'h': 1.0, 'x': 0.0, 'w': 1.0, 's': 1.1, 'e': 1.0}
        for v in def_vals:
            if v not in kw:
                kw[v] = def_vals[v]

        # this copies all of the defaults into what the peak function needs
        anames = list(inspect.signature(f).parameters)
        fkw = dict([(arg, kw[arg]) for arg in anames if arg in kw])

        # some functions use location or width parameters explicitly
        # if not, adjust the timeseries accordingly
        ta = t
        if 'x' not in anames:
            ta = ta - kw['x']
        if 'w' not in anames:
            ta = ta / kw['w']

        # finally call the function
        mod = f(ta, **fkw)
        # recalcualte, making the peak maximize at x
        mod = f(ta + ta[mod.argmax()], **fkw)
        return kw['v'] + kw['h'] / max(mod) * mod

    args = set(['v', 'h', 'x', 'w'])
    anames = list(inspect.signature(f).parameters)
    wrapped_f._peakargs = list(args.union([a for a in anames
                                           if a not in ('t', 'r')]))
    return wrapped_f


@peak_model
def bigaussian(t, w, s):
    # for an example of use: http://www.biomedcentral.com/1471-2105/11/559
    # Di Marco & Bombi use formulation with w1 & w2, but it
    # looks better to use formulation with w and s
    w1, w2 = w * exp(-s) / (1 + exp(-s)), w / (1 + exp(-s))
    y = np.empty(len(t))
    y[t < 0] = exp(-t[t < 0] ** 2 / (2 * w1 ** 2)) / sqrt(2 * np.pi)
    y[t >= 0] = exp(-t[t >= 0] ** 2 / (2 * w2 ** 2)) / sqrt(2 * np.pi)
    return y


@peak_model
def box(t):
    y = np.zeros(len(t))
    y[np.logical_and(t > -0.5, t < 0.5)] = 1.0
    return y


@bounds(w=(openlow(0.), np.inf), s=(1., np.inf))
@peak_model
def exp_mod_gaussian(t, w, s):
    # http://en.wikipedia.org/wiki/Exponentially_modified_Gaussian_distribution
    exp_t = exp((w ** 2 - 2 * s * t) / (2 * s ** 2))
    erf_t = erfc((w ** 2 - s * t) / (s * w))
    return (w ** 1.5) / (1.414214 * s) * exp_t * erf_t


@peak_model
def extreme_value(t):
    return exp(-exp(-t) - t + 1)


@bounds(w=(openlow(0.), np.inf), s=(1., np.inf))
@peak_model
def gamma_dist(t, w, s):
    # from Wikipedia: not the same as Di Marco & Bombi's formulation
    # s > 1
    y = np.zeros(len(t))
    y[t > 0] = t[t > 0] ** (s - 1) * exp(-t[t > 0] / w) / (w ** s * gamma(s))
    return y


@peak_model
def gaussian(t):
    """
    Gaussian
    """
    return exp(-0.5 * t ** 2)


@bounds(w=(openlow(0.), np.inf), x=(openlow(0.), np.inf))
@peak_model
def giddings(t, w, x):
    print(w, x)
    # w != 0
    y = np.zeros(len(t))
    y[t > 0] = (1. / w) * sqrt(x / t[t > 0]) * exp((t[t > 0] + x) / -w)
    # TODO: "overflow encountered in i1"
    # y[t > 0] *= i1(2. * sqrt(x * t[t > 0]) / w)
    # trying to keep the shape, but not allow such high numbers?
    y[t > 0] *= i1(np.linspace(2, 10, sum(t > 0)))
    return y


@bounds(w=(openlow(0.1), np.inf), s=(openlow(0.), np.inf))
@peak_model
def haarhoffvanderlinde(t, w, s):
    # s here = s * z in Di Marco & Bombi
    # w, s != 0
    y = w * exp(-0.5 * (t / w) ** 2) / (s * sqrt(2 * np.pi))
    print(s, w)
    y /= 1. / (exp(s / w ** 2) - 1.) + 0.5 * erfc(t / (w * sqrt(2)))
    return y


@bounds(w=(openlow(0.), np.inf), s=(openlow(1.), np.inf))
@peak_model
def lognormal(t, w, s, r=2.):
    # r is the ratio between h and the height at
    # which s is computed: normally 2.
    y = np.zeros(len(t))
    # TODO: if log(s) rounds to 0, big problems here
    lt = -log(r) / log(s) ** 2
    # try to adjust timing so peak stays centered at 0
    ta = t + (w - 1) / (1.12383 * s - 0.780647)
    y[ta > 0] = exp(lt * log(ta[ta > 0] / w * (s ** 2 - 1) / s) ** 2)
    return y


@peak_model
def lorentzian(t, a):
    # from Wikipedia: not the same as Di Marco & Bombi's formulation
    # return 1. / (1. + 4. * t ** 2)
    return a / (np.pi * (a ** 2 + t ** 2))


@peak_model
def papai_pap(t, s, e):
    # s is skewness, e is excess
    y = np.zeros(len(t))
    ft = t[t > 0]
    y[t > 0] = 1 + (s / 6.) * (ft ** 2 - 3. * ft)
    y[t > 0] -= (e / 24.) * (ft ** 4 - 6 * ft ** 2 + 3)
    y[t > 0] *= exp(-0.5 * ft)
    return y


@peak_model
def parabola(t):
    y = np.zeros(len(t))
    mask = np.logical_and(t > -1, t < 1)
    y[mask] = 1 - t[mask] ** 2
    return y


@bounds(a=(openlow(0.), np.inf))
@peak_model
def pearson_vii(t, a):
    return (1 + 4 * t ** 2 * (2 ** (1 / a) - 1)) ** -a


@peak_model
def poisson(t, a):
    # a > 1
    y = np.zeros(len(t))
    y[t > 0] = exp((1 - a) * (t[t > 0] - log(t[t > 0]) - 1))
    return y


@bounds(s=(openlow(0.), np.inf))
@peak_model
def studentt(t, s):
    # s != 0
    return (1 + (t ** 2) / s) ** (-0.5 * (s + 1))


@peak_model
def triangle(t):
    y = np.zeros(len(t))
    mask = np.logical_and(t > -0.5, t < 0.5)
    y[mask] = 1 - abs(t[mask]) / 0.5
    return y


@bounds(a=(openlow(1.), np.inf))
@peak_model
def weibull3(t, a):
    # TODO: doesn't work?
    y = np.zeros(len(t))
    at = (a - 1.) / a
    tt = t[t > 0] + ((a - 1.) / a) ** (1. / a)
    y[t > 0] = at ** at * tt ** (a - 1.) * exp(-tt ** a + at)
    return y

#  FUNCTIONS TO DO
# def chesler_cram_a(t, a, b, c, d):
# def chesler_cram_b(t, a, b, c, d, e):
# def cumulative(t, w, a):
# def f_variance(t, s1, s2):
# def gladney_dowden_a(t, w, s):
# def gladney_dowden_b(t, w, a, b):
# def haldna_phi(t, w, s):
# def intermediate(t, a, b):
# def li_a(t, w):
# def li_b(t, w, a1, a2, b1, b2):
# def losev(t, w1, w2, a):
# def nonlinearchromatography(t, x, w, s):
#     # also r & v?
# def pearsonIV(t, w, s1, s2):
# def pearsonIVa(t, w, s):
# def pearsonIVb(t, w, s):
# def pseudovoight1(t, a):
# def pseudovoight2(t, s, a):
# def pulse(t, a):
#
#
# def gram_charlier(t, w, *n):
#     # TODO: implement this; Berberan-Santos '07 has
#     # ways to calculate cumulant values
#     raise NotImplementedError
#
#
# def edgeworth_cramer(t, *n):
#     # y = exp(-0.5 * t ** 2)
#     raise NotImplementedError


peak_models = [bigaussian, box, exp_mod_gaussian, extreme_value, gamma_dist,
               gaussian, giddings, haarhoffvanderlinde, lognormal, lorentzian,
               papai_pap, parabola, pearson_vii, poisson, studentt, triangle,
               weibull3]



import multiprocessing
import functools
import numpy as np
from aston.trace.trace import Trace
from aston.peak.peak import Peak, PeakComponent
from aston.peak.peak_models import peak_models
from aston.peak.peak_fitting import guess_initc, fit


def simple_integrate(ts, peak_list, base_ts=None, intname='simple'):
    """
    Integrate each peak naively; without regard to overlap.

    This is used as the terminal step by most of the other integrators.
    """
    peaks = []
    for hints in peak_list:
        t0, t1 = hints['t0'], hints['t1']
        hints['int'] = intname
        pk_ts = ts.twin((t0, t1))
        if base_ts is None:
            # make a two point baseline
            base = Trace([hints.get('y0', pk_ts[0]),
                          hints.get('y1', pk_ts[-1])],
                         [t0, t1], name=ts.name)
        else:
            base = base_ts.twin((t0, t1))
        peaks.append(PeakComponent(hints, pk_ts, base))
    return peaks


def _get_windows(peak_list):
    """
    Given a list of peaks, bin them into windows.
    """
    win_list = []
    for hints in peak_list:
        p_w = hints['t0'], hints['t1']
        for w in win_list:
            if p_w[0] <= w[0][1] and p_w[1] >= w[0][0]:
                w[0] = (min(p_w[0], w[0][0]), max(p_w[1], w[0][1]))
                w[1].append(hints)
                break
        else:
            win_list.append([p_w, [hints]])
    return win_list


def constant_bl_integrate(ts, peak_list):
    temp_pks = []
    for hints in peak_list:
        min_y = np.min(ts.twin((hints['t0'], hints['t1'])).values)
        prop = hints.copy()
        prop['y0'], prop['y1'] = min_y, min_y
        temp_pks.append(prop)
    peaks = simple_integrate(ts, temp_pks, intname='constant_bl')
    return peaks


def drop_integrate(ts, peak_list):
    """
    Resolves overlap by breaking at the minimum value.
    """
    peaks = []
    for _, pks in _get_windows(peak_list):
        temp_pks = []
        pks = sorted(pks, key=lambda p: p['t0'])
        if 'y0' in pks[0] and 'y1' in pks[-1]:
            y0, y1 = pks[0]['y0'], pks[-1]['y1']
        else:
            y0 = ts.get_point(pks[0]['t0'])
            y1 = ts.get_point(pks[-1]['t1'])
        ys = np.array([y0, y1])
        xs = np.array([pks[0]['t0'], pks[-1]['t1']])

        # go through list of peaks to make sure there's no overlap
        for hints in pks:
            t0, t1 = hints['t0'], hints['t1']

            # figure out the y values (using a linear baseline)
            hints['y0'] = np.interp(t0, xs, ys)
            hints['y1'] = np.interp(t1, xs, ys)

            # if this peak totally overlaps with an existing one, don't add
            if sum(1 for p in temp_pks if t1 <= p['t1']) > 0:
                continue
            overlap_pks = [p for p in temp_pks if t0 <= p['t1']]
            if len(overlap_pks) > 0:
                # find the last of the overlapping peaks
                overlap_pk = max(overlap_pks, key=lambda p: p['t0'])
                # get the section of trace and find the lowest point
                over_ts = ts.twin((t0, overlap_pk['t1']))
                min_t = over_ts.index[over_ts.values.argmin()]

                # delete the existing overlaping peak
                for i, p in enumerate(temp_pks):
                    if p == overlap_pk:
                        del temp_pks[i]
                        break

                # interpolate a new y value
                y_val = np.interp(min_t, xs, ys)
                overlap_pk['y1'] = y_val
                hints['y0'] = y_val

                # add the old and new peak in
                overlap_pk['t1'] = min_t
                temp_pks.append(overlap_pk)
                hints['t0'], hints['t1'] = min_t, t1
                temp_pks.append(hints)
            else:
                hints['t0'], hints['t1'] = t0, t1
                temp_pks.append(hints)

        # none of our peaks should overlap, so we can just use
        # simple_integrate now
        peaks += simple_integrate(ts, temp_pks, intname='drop')
    return peaks


def leastsq_integrate(ts, peak_list, f='gaussian'):
    # FIXME: transition from t0, t1, params to params (containing 't0'/'t1'
    # lookup the peak model function to use for fitting
    f = {f.__name__: f for f in peak_models}[f]

    peaks = []
    for w, peak_list in _get_windows(peak_list):
        tr = ts.trace(twin=w)
        # if there's location info, we can use this for
        # our inital params
        if 'x' in peak_list[0][2]:
            # TODO: should fill in other values?
            # TODO: should do something with y0 and y1?
            initc = [i[2] for i in peak_list]
        else:
            # TODO: should find peak maxima for rts?
            rts = [0.5 * (i[0] + i[1]) for i in peak_list]
            initc = guess_initc(tr, f, rts)
        params, _ = fit(tr, [f] * len(initc), initc)
        for p in params:
            pk_ts = Trace(f(tr.times, **p), tr.times)
            info = {'name': '{:.2f}'.format(p['x'])}
            info['p-create'] = '{},leastsq_integrate'.format(
                peak_list[0][2].get('pf', '')
            )
            pk = PeakComponent(info, pk_ts)
            peaks.append(pk)
    return peaks


def periodic_integrate(ts, peak_list, offset=0., period=1.):
    # TODO: should be a peak finder, not an integrator?
    def movwin(a, length):
        return np.lib.stride_tricks.as_strided(
            a,
            (a.shape[0] - length + 1, length),
            a.itemsize * np.ones(2)
        )
    new_peak_list = []
    for hints in peak_list:
        t0, t1 = hints['t0'], hints['t1']

        # time the first whole "period" starts
        tpi = offset + period * ((t0 - offset) // period + 1)
        if tpi > t1:
            # the entire peak is within one "period"
            new_peak_list.append([t0, t1, hints])
            continue
        tp = np.hstack([[t0], np.arange(tpi, t1, period)])
        if tp[-1] != t1:
            # add the last point to the list
            tp = np.hstack([tp, [t1]])
        for tp0, tp1 in movwin(tp, 2):
            new_hints = {'pf': hints.get('pf', '')}
            if 'y0' in hints and 'y1' in hints:
                # calculate the new baseline for this peak
                xs, ys = [t0, t1], [hints['y0'], hints['y1']]
                new_hints['y0'] = np.interp(tp0, xs, ys)
                new_hints['y1'] = np.interp(tp1, xs, ys)
            new_peak_list.append([tp0, tp1, new_hints])

    peaks = simple_integrate(ts, new_peak_list)
    for p in peaks:
        p.info['p-create'] = '{},periodic_integrate'.format(
            p.info['p-create'].split(',')[0]
        )
    return peaks


def _peakcomp_to_name(pkc):
    return '{:.3f}-{:.3f}'.format(pkc.info['t0'], pkc.info['t1'])


def merge_peaks_by_time(pks_list, time_diff=0.01):
    sort_pks = sorted([pk for pks in pks_list for pk in pks],
                      key=lambda pk: pk.time())

    # add the first component in as a peak
    pkc = sort_pks[0]
    pk = Peak(name=_peakcomp_to_name(pkc), components=[pkc])
    mrg_pks = [pk]

    for pkc in sort_pks[1:]:
        c_pk = mrg_pks[-1]
        if np.abs(c_pk.components[0].time() - pkc.time()) < time_diff and \
           pkc._trace.name not in {c._trace.name for c in c_pk.components}:
            c_pk.components.append(pkc)
        else:
            pk = Peak(name=_peakcomp_to_name(pkc), components=[pkc])
            mrg_pks.append(pk)
    return mrg_pks


def merge_peaks_by_order(pks_list):
    """

    All lists of peaks need to be the same length.
    """
    mrg_pks = []
    for sub_pks in zip(*pks_list):
        pk = Peak(name=_peakcomp_to_name(sub_pks[0]), components=sub_pks)
        mrg_pks.append(pk)
    return mrg_pks


def _integrate_mpwrap(ts_and_pks, integrate, fopts):
    """
    Take a zipped timeseries and peaks found in it
    and integrate it to return peaks. Used to allow
    multiprocessing support.
    """
    ts, tpks = ts_and_pks
    pks = integrate(ts, tpks, **fopts)
    # for p in pks:
    #     p.info['mz'] = str(ts.name)
    return pks


def integrate_peaks(tss, peaks_found, int_f, f_opts={},
                    as_first=False, mp=False):
    f = functools.partial(_integrate_mpwrap, integrate=int_f, fopts=f_opts)
    if mp:
        po = multiprocessing.Pool()
        all_pks = po.map(f, zip(tss, peaks_found))
        po.close()
        po.join()
    else:
        all_pks = list(map(f, zip(tss, peaks_found)))

    # merge peaks from all_pks together
    if as_first:
        mrg_pks = merge_peaks_by_order(all_pks)
    else:
        mrg_pks = merge_peaks_by_time(all_pks)
    return mrg_pks



# -*- coding: utf-8 -*-
import numpy as np
import re
from aston.trace import Trace
from aston.trace.math_chromatograms import molmz, mzminus, basemz
from aston.trace.math_traces import (ts_func, fft, noisefilter,
                                     movingaverage, savitzkygolay)

functions = {}  # TODO
istr_type_2d = ['ms', 'uv', 'irms']
istr_type_evts = ['refgas', 'fia', 'fxn', 'deadvol']

SHORTCUTS = {
    'alkanes': '57#ms+71#ms+85#ms',
    'alkenes': '55#ms+69#ms+83#ms',
    'alkadienes': '67#ms+81#ms',
    'hopanes': '191#ms',
    'hopenes': '189#ms',
    'steranes': '217#ms',
    'ketones': '59#ms',
    'amides': '59#ms',
    'tms': '73#ms',
    'methylesters': '74#ms',
    'alcohols': '75#ms',
    'pthalates': '149#ms'
}


def tokens(istr):
    """
    Same as tokenize, but returns only tokens
    (and at all parantheses levels).
    """

    # make a list of all alphanumeric tokens
    toks = re.findall(r'[^\*\\\+\-\^\(\)]+\(?', istr)

    # remove the functions
    return [t for t in toks if not t.endswith('(')]


def istr_type(istr):
    """
    Given an "ion" specification, determine its "type", e.g. 1D, Events, etc.
    """
    data = set(i.rstrip('0123456789') for i in tokens(istr))
    has_events = not data.isdisjoint(istr_type_evts)
    has_2d = not data.isdisjoint(istr_type_2d)
    has_1d = data.difference(istr_type_evts).difference(istr_type_2d) != set()

    if has_events and not (has_1d or has_2d):
        return 'events'
    elif has_1d and not has_events:
        return '1d'
    elif has_2d and not (has_events or has_1d):
        return '2d'
    else:
        return None


def istr_best_2d_source(istr, avail_sources=None):
    if avail_sources is None:
        avail_sources = []
    for token in tokens(istr):
        s = token.split('#')[-1]
        if s.rstrip('0123456789') in istr_type_2d and s in avail_sources:
            return s
    for s in istr_type_2d:
        if s in avail_sources:
            return s
    return None


def token_source(token, avail_sources=None):
    prim_2d = None
    if avail_sources is None:
        avail_sources = []
    else:
        # find the "best" 2d trace available
        for s in istr_type_2d:
            if s in avail_sources:
                prim_2d = s
                break

    if token in avail_sources:
        if token.rstrip('0123456789') in istr_type_2d:
            return 'tic', token
        else:
            return token, token
    elif '#' in token:
        return token.rsplit('#', 1)
    elif token in {'tic', 'x', ''} and 'fid' in avail_sources:
        return token, 'fid'
    else:
        return token, prim_2d


def tokenize(istr, token):
    plevel, words, sacc = 0, [], ''
    for s in istr.split(token):
        plevel += sum(1 for i in s if i == '(')
        plevel -= sum(1 for i in s if i == ')')
        if sacc == '':
            sacc += s
        else:
            sacc += token + s
        if plevel == 0:
            words.append(sacc.strip())
            sacc = ''
    if sacc != '':
        words.append(sacc.strip())
    return words


def is_parans_exp(istr):
    """
    Determines if an expression is a valid function "call"
    """
    fxn = istr.split('(')[0]
    if (not fxn.isalnum() and fxn != '(') or istr[-1] != ')':
        return False
    plevel = 1
    for c in '('.join(istr[:-1].split('(')[1:]):
        if c == '(':
            plevel += 1
        elif c == ')':
            plevel -= 1
        if plevel == 0:
            return False
    return True


def trace_resolver(istr, analyses, twin=None):
    avail_sources = [i.lstrip('#*') for a in analyses
                     for i in a.trace.split(',')]
    istr, source = token_source(istr, avail_sources)
    if source is None:
        return Trace()

    for a in analyses:
        if source in [i.lstrip('#*') for i in a.trace.split(',')]:
            df = a.datafile
            break
    else:
        df = None

    if istr in {'coda', 'rnie', 'wmsm'}:
        # TODO: allow more complicated options to turn
        # Chromatograms into plotable Traces

        # coda
        #  Windig W: The use of the Durbin-Watson criterion for
        #  noise and background reduction of complex liquid
        #  chromatography/mass spectrometry data and a new algorithm
        #  to determine sample differences. Chemometrics and
        #  Intelligent Laboratory Systems. 2005, 77:206-214.

        # rnie
        #  Yunfei L, Qu H, and Cheng Y: A entropy-based method
        #  for noise reduction of LC-MS data. Analytica Chimica
        #  Acta 612.1 (2008)

        # wmsm
        #  Fleming C et al. Windowed mass selection method:
        #  a new data processing algorithm for LC-MS data.
        #  Journal of Chromatography A 849.1 (1999) 71-85.
        pass
    elif istr.startswith('m_'):
        if istr == 'm_':
            m = 0.0
        else:
            m = float(istr.split('_')[1])
        return mzminus(df.data, m)
    elif istr == 'molmz':
        return molmz(df.data)
    elif istr == 'basemz':
        return basemz(df.data)
    elif istr in {'r45std', 'r46std'}:
        # TODO: calculate isotopic data -> needs integrated peak
        # references of associated peaks in order to make these calculations
        pass
        #  calculate isotopic reference for chromatogram
        # if name == 'r45std':
        #     topion = 45
        # else:
        #     topion = 46
        # std_specs = [o for o in \
        #   self.children_of_type('peak') \
        #   if o.info['p-type'] == 'Isotope Standard']
        # x = [float(o.info['p-s-time']) for o in std_specs]
        # y = [o.area(topion) / o.area(44) for o in std_specs \
        #      if o.area(44) != 0]
        # if len(x) == 0 or len(y) == 0:
        #     return self._const(0.0, twin)

        # p0 = [y[0], 0]
        # errfunc = lambda p, x, y: p[0] + p[1] * x - y
        # try:
        #     p, succ = leastsq(errfunc, p0, args=(np.array(x), \
        #                                          np.array(y)))
        # except:
        #     p = p0
        # sim_y = np.array(errfunc(p, t, np.zeros(len(t))))
        # return TimeSeries(sim_y, t, [name])
    else:
        # interpret tolerances
        if ':' in istr:
            st = float(istr.split(':')[0])
            en = float(istr.split(':')[1])
            tol = 0.5 * (en - st)
            istr = 0.5 * (en + st)
        elif u'±' in istr:
            tol = float(istr.split(u'±')[1])
            istr = float(istr.split(u'±')[0])
        else:
            tol = 0.5

        return df.trace(istr, tol, twin=twin)


def fxn_resolver():
    # FIXME: not used properly (or at all)
    fxns = {
        'fft': fft,
        'noise': ts_func(noisefilter),
        'abs': ts_func(np.abs),
        'sin': ts_func(np.sin),
        'cos': ts_func(np.cos),
        'tan': ts_func(np.tan),
        'derivative': ts_func(np.gradient),
        'd': ts_func(np.gradient),
        'movingaverage': movingaverage,
        'savitzkygolay': savitzkygolay,
    }
    return fxns


def parse_ion_string(istr, analyses, twin=None):
    """
    Recursive string parser that handles "ion" strings.
    """

    if istr.strip() == '':
        return Trace()

    # remove (unnessary?) pluses from the front
    # TODO: plus should be abs?
    istr = istr.lstrip('+')

    # invert it if preceded by a minus sign
    if istr[0] == '-':
        return -parse_ion_string(istr[1:], analyses, twin)

    # this is a function or paranthesized expression
    if is_parans_exp(istr):
        if ')' not in istr:
            # unbalanced parantheses
            pass
        fxn = istr.split('(')[0]
        args = istr[istr.find('(') + 1:istr.find(')')].split(',')
        if fxn == '':
            # strip out the parantheses and continue
            istr = args[0]
        else:
            ts = parse_ion_string(args[0], analyses, twin)
            # FIXME
            return ts
            # return fxn_resolver(ts, fxn, *args[1:])

    # all the complicated math is gone, so simple lookup
    if set(istr).intersection(set('+-/*()')) == set():
        if istr in SHORTCUTS:
            # allow some shortcuts to pull out common ions
            return parse_ion_string(SHORTCUTS[istr], analyses, twin)
        elif istr[0] == '!' and all(i in '0123456789.' for i in istr[1:]):
            # TODO: should this handle negative numbers?
            return float(istr[1:])
        elif istr == '!pi':
            return np.pi
        elif istr == '!e':
            return np.e
        else:
            return trace_resolver(istr, analyses, twin)

    # go through and handle operators
    for token in '/*+-^':
        if len(tokenize(istr, token)) != 1:
            ts = tokenize(istr, token)
            s = parse_ion_string(ts[0], analyses, twin)
            for t in ts[1:]:
                if token == '/':
                    s /= parse_ion_string(t, analyses, twin)
                elif token == '*':
                    s *= parse_ion_string(t, analyses, twin)
                elif token == '+':
                    s += parse_ion_string(t, analyses, twin)
                elif token == '-':
                    s -= parse_ion_string(t, analyses, twin)
                elif token == '^':
                    s **= parse_ion_string(t, analyses, twin)
            return s
    raise Exception('Parser hit a point it shouldn\'t have!')



import numpy as np
import struct
from aston.trace import Trace
from aston.tracefile import TraceFile


class AgilentFID(TraceFile):
    # TODO: preliminary support, math may not be correct
    """
    Reads a Agilent FID *.ch file.
    """
    mime = 'application/vnd-agilent-chemstation-fid'
    traces = ['fid']

    def _trace(self, name, twin):
        # this method allows self.trace('fid') to get something
        return self.total_trace(twin)

    def total_trace(self, twin=None):
        f = open(self.filename, 'rb')

        f.seek(0x11A)
        start_time = struct.unpack('>f', f.read(4))[0] / 60000.
        # end_time = struct.unpack('>f', f.read(4))[0] / 60000.
        # end_time 0x11E '>i'

        # FIXME: why is there this del_ab code here?
        # f.seek(0x284)
        # del_ab = struct.unpack('>d', f.read(8))[0]
        data = []

        f.seek(0x400)
        delt = 0
        while True:
            try:
                inp = struct.unpack('>h', f.read(2))[0]
            except struct.error:
                break

            if inp == 32767:
                inp = struct.unpack('>i', f.read(4))[0]
                inp2 = struct.unpack('>H', f.read(2))[0]
                delt = 0
                data.append(inp * 65534 + inp2)
            else:
                delt += inp
                data.append(data[-1] + delt)
        f.close()
        # TODO: 0.4/60.0 should be obtained from the file???
        times = np.array(start_time + np.arange(len(data)) * (0.2 / 60.0))
        # times = np.linspace(start_time, end_time, data.shape[0])
        return Trace(np.array([data]).T, times, name='TIC')


class AgilentFID2(TraceFile):
    # TODO: preliminary support, math may not be correct
    """
    Reads a Agilent FID *.ch file.
    """
    mime = 'application/vnd-agilent-chemstation-fid2'
    traces = ['fid']

    def _trace(self, name, twin):
        # this method allows self.trace('fid') to get something
        return self.total_trace(twin)

    def total_trace(self, twin=None):
        f = open(self.filename, 'rb')

        f.seek(0x11A)
        start_time = struct.unpack('>f', f.read(4))[0] / 60000.
        end_time = struct.unpack('>f', f.read(4))[0] / 60000.

        # TODO: figure out if this exists and where?
        # FID signal seems like 10x higher than it should be?
        # f.seek(0x284)
        # del_ab = 0.1  # struct.unpack('>d', f.read(8))[0]
        # data = []

        f.seek(0x1800)
        data = np.fromfile(f, '<f8')
        times = np.linspace(start_time, end_time, data.shape[0])
        return Trace(data, times, name='TIC')



import struct
import json
import zlib
import numpy as np
import scipy.io.wavfile
import scipy.signal
import scipy.sparse
from scipy.interpolate import interp1d
from aston.spectra import Scan


class Trace(object):
    def __init__(self, data, index=None, name=''):
        # TODO: reenable this without having to import from pandas
        # if isinstance(data, Series):
        #     self.values = data.values
        #     self.index = data.index.values
        #     self.name = data.name
        if index is not None:
            self.values = np.array(data)
            self.index = np.array(index)
            self.name = name
        else:
            # need either one pandas.Series or two np.arrays
            raise TypeError('Trace initialized improperly.')

    @property
    def shape(self):
        return self.values.shape

    def copy(self):
        return Trace(self.values.copy(), self.index.copy(), self.name)

    def __len__(self):
        return self.index.shape[0]

    def __getitem__(self, index):
        v, i = self.values[index], self.index[index]
        if type(v) is np.ndarray:
            return Trace(v, i, self.name)
        else:
            return v

    def plot(self, style='-', color='k', scale=False, label=None, ax=None):
        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()

        if scale:
            # normalize data to 0 to 1
            y = (self.values - np.min(self.values)) / np.max(self.values)
        else:
            y = self.values

        if label is None:
            label = self.name

        ax.plot(self.index, y, c=color, ls=style, label=label)

    def twin(self, twin):
        st_idx, en_idx = _slice_idxs(self, twin)
        return self[st_idx:en_idx]

    def as_text(self, width=80, height=20):
        raise NotImplementedError
        # break s.index into width chunks
        # find max and min of each chunk along with start and stop?

    def get_point(self, time, interp_method=None):
        # TODO: add more interpolation methods
        if len(self.index) < 2:
            return np.nan
        f = interp1d(self.index, self.values,
                     bounds_error=False, fill_value=0.0)
        return f(time)

    def adjust_time(self, offset=0.0, scale=1.0):
        adjs = self.copy()
        adjs.index = adjs.index * scale + offset
        adjs.__class__ = Trace
        return adjs

    def _retime(self, new_times, fill=0.0):
        # this is not exposed because it returns a raw numpy array
        from scipy.interpolate import interp1d
        if new_times.shape == self.shape[0]:
            if np.all(np.equal(new_times, self.index)):
                return self

        def f(d):
            return interp1d(self.index, d,
                            bounds_error=False, fill_value=fill)(new_times)
        return np.apply_along_axis(f, 0, self.values)

    def _apply_data(self, f, ts, reverse=False):
        """
        Convenience function for all of the math stuff.
        """
        # TODO: needs to catch np numeric types?
        if isinstance(ts, (int, float)):
            d = ts * np.ones(self.shape[0])
        elif ts is None:
            d = None
        elif np.array_equal(ts.index, self.index):
            d = ts.values
        else:
            d = ts._retime(self.index)

        if not reverse:
            new_data = np.apply_along_axis(f, 0, self.values, d)
        else:
            new_data = np.apply_along_axis(f, 0, d, self.values)
        return Trace(new_data, self.index, name=self.name)

    def __add__(self, ts):
        return self._apply_data(lambda x, y: x + y, ts)

    def __sub__(self, ts):
        return self._apply_data(lambda x, y: x - y, ts)

    def __mul__(self, ts):
        return self._apply_data(lambda x, y: x * y, ts)

    def __div__(self, ts):
        return self._apply_data(lambda x, y: x / y, ts)

    def __truediv__(self, ts):
        return self.__div__(ts)

    def __reversed(self):
        raise NotImplementedError

    # TODO: this should happen in place?
    def __iadd__(self, ts):
        return self._apply_data(lambda x, y: x + y, ts)

    def __isub__(self, ts):
        return self._apply_data(lambda x, y: x - y, ts)

    def __imul__(self, ts):
        return self._apply_data(lambda x, y: x * y, ts)

    def __idiv__(self, ts):
        return self._apply_data(lambda x, y: x / y, ts)

    def __radd__(self, ts):
        return self._apply_data(lambda x, y: x + y, ts, reverse=True)

    def __rsub__(self, ts):
        return self._apply_data(lambda x, y: x - y, ts, reverse=True)

    def __rmul__(self, ts):
        return self._apply_data(lambda x, y: x * y, ts, reverse=True)

    def __rdiv__(self, ts):
        return self._apply_data(lambda x, y: x / y, ts, reverse=True)

    def __neg__(self):
        return self._apply_data(lambda x, y: -x, None)

    def __abs__(self):
        return self._apply_data(lambda x, y: abs(x), None)

    def compress(self):
        i = self.index.astype(np.float32).tostring()
        li = struct.pack('<L', len(i))
        c = json.dumps([self.name]).encode('utf-8')
        lc = struct.pack('<L', len(c))
        v = self.values.astype(np.float64).tostring()
        try:  # python 2
            return buffer(zlib.compress(lc + li + c + i + v))
        except NameError:  # python 3
            return zlib.compress(lc + li + c + i + v)


class Chromatogram(object):
    def __init__(self, data=None, index=None, columns=None, yunits=None):
        self.yunits = yunits
        if data is None:
            self.values = np.array([])
            self.index = np.array([])
            self.columns = ['']
        # TODO: reenable this without having to import from pandas
        # elif isinstance(data, DataFrame):
        #     self.values = data.values
        #     self.index = data.index.values
        #     self.columns = data.columns.values
        elif index is not None:
            # TODO: handle sparse arrays
            if isinstance(data, list):
                self.values = np.array(data)
            else:
                self.values = data
            self.index = np.array(index)
            self.columns = columns
        else:
            # need either one pandas.Series or two np.arrays
            raise TypeError('Chromatogram initialized improperly.')

    @property
    def shape(self):
        return self.values.shape

    def copy(self):
        return Chromatogram(self.values.copy(), self.index.copy(),
                            self.columns.copy())

    def __len__(self):
        return self.index.shape[0]

    def __getitem__(self, key):
        if isinstance(key, tuple):
            # indexing on multiple dimensions
            # TODO: check that dimensions are in right order
            v = self.values[key]
            i = self.index[key[0]]
            c = self.columns[key[1]]
        else:
            if isinstance(key, np.ndarray):
                if len(key.shape) > 1:
                    # indexing with a boolean mask
                    raise NotImplementedError
            v, i = self.values[key], self.index[key]
            c = self.columns

        if len(c) == 1:
            if len(i) == 1:
                return v[0, 0]
            else:
                return Trace(v, i, name=c[0])
        else:
            return Chromatogram(v, i, c)

    @property
    def traces(self):
        """
        Decomposes the Chromatogram into a collection of Traces.

        Returns
        -------
        list
        """
        traces = []
        for v, c in zip(self.values.T, self.columns):
            traces.append(Trace(v, self.index, name=c))
        return traces

    def trace(self, name='tic', tol=0.5, twin=None):
        # TODO: aggfunc in here for tic and numeric
        st_idx, en_idx = _slice_idxs(self, twin)

        if isinstance(name, (int, float, np.float32, np.float64)):
            name = str(name)

        if name in ['tic', 'x', '']:
            data = self.values.sum(axis=1)
            name = 'tic'
        elif name == '!':
            data = self[:, 0]
            name = self.columns[0]
        elif set(name).issubset('1234567890.'):
            cols = np.genfromtxt(np.array(self.columns).astype(bytes))
            cols = np.abs(cols - float(name)) < tol
            if not np.any(cols):
                data = np.zeros(self.shape[0]) * np.nan
            elif isinstance(self.values, scipy.sparse.spmatrix):
                data = self.values[:, cols].toarray().sum(axis=1)
            else:
                data = self.values[:, cols].sum(axis=1)
        else:
            data = np.zeros(self.shape[0]) * np.nan
            name = ''

        # TODO: better way to coerce this into the right class?
        # TODO: use twin
        return Trace(data, index=self.index, name=name)

    def plot(self, style='heatmap', legend=False, cmap=None, ax=None):
        """
        Presents the AstonFrame using matplotlib.

        Parameters
        ----------
        style : {'heatmap', 'colors', ''}
        legend : bool, optional
        cmap: matplotlib.colors.Colormap, optional
        ax : matplotlib.axes.Axes, optional

        """
        # styles: 2d, colors, otherwise interpret as trace?
        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()

        if style == 'heatmap':
            ions = self.columns
            ext = (self.index[0], self.index[-1], min(ions), max(ions))
            grid = self.values[:, np.argsort(self.columns)].transpose()
            if isinstance(self.values, scipy.sparse.spmatrix):
                grid = grid.toarray()
            img = ax.imshow(grid, origin='lower', aspect='auto',
                            extent=ext, cmap=cmap)
            if legend:
                ax.figure.colorbar(img)
        elif style == 'colors':
            # TODO: importing gaussian at the top leads to a whole
            # mess of dependency issues => fix somehow?
            from aston.peak.peak_models import gaussian
            from matplotlib.colors import ListedColormap

            wvs = np.genfromtxt(np.array(self.columns).astype(bytes))
            # wvs = self.columns.astype(float)

            # http://www.ppsloan.org/publications/XYZJCGT.pdf
            vis_filt = np.zeros((3, len(wvs)))
            vis_filt[0] = 1.065 * gaussian(wvs, x=595.8, w=33.33) + \
                0.366 * gaussian(wvs, x=446.8, w=19.44)
            vis_filt[1] = 1.014 * gaussian(np.log(wvs), x=np.log(556.3),
                                           w=0.075)
            vis_filt[2] = 1.839 * gaussian(np.log(wvs), x=np.log(449.8),
                                           w=0.051)
            if isinstance(self.values, scipy.sparse.spmatrix):
                xyz = np.dot(self.values.toarray(), vis_filt.T)
            else:
                xyz = np.dot(self.values.copy(), vis_filt.T)

            # http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
            xyz_rgb = [[3.2404542, -1.5371385, -0.4985314],
                       [-0.9692660, 1.8760108, 0.0415560],
                       [0.0556434, -0.2040259, 1.0572252]]
            xyz_rgb = np.array(xyz_rgb)
            rgb = np.dot(xyz_rgb, xyz.T).T

            # normalize
            rgb[rgb < 0] = 0
            rgb /= np.max(rgb)
            rgb = 1 - np.abs(rgb)

            # plot
            cmask = np.meshgrid(np.arange(rgb.shape[0]), 0)[0]
            ax.imshow(cmask, cmap=ListedColormap(rgb), aspect='auto',
                      extent=(self.index[0], self.index[-1], 0, 1))
            ax.yaxis.set_ticks([])
        else:
            if cmap is not None:
                color = cmap(0, 1)
            else:
                color = 'k'
            self.trace().plot(color=color, ax=ax)

    def as_sound(self, filename, speed=60, cutoff=50):
        """
        Convert AstonFrame into a WAV file.

        Parameters
        ----------
        filename : str
            Name of wavfile to create.
        speed : float, optional
            How much to speed up for sound recording, e.g. a value of 60
            will turn an hour-long AstonFrame into a minute-long sound clip.
        cutoff : float, optional
            m/z's under this value will be clipped out.
        """
        # make a 1d array for the sound
        def to_t(t):
            return (t - self.index[0]) / speed

        wav_len = int(to_t(self.index[-1]) * 60 * 44100)
        wav = np.zeros(wav_len)

        # create an artificial array to interpolate times out of
        tmask = np.linspace(0, 1, self.shape[0])

        # come up with a mapping from mz to tone
        min_hz, max_hz = 50, 1000
        min_mz, max_mz = min(self.columns), max(self.columns)

        def mz_to_wv(mz):
            """
            Maps a wavelength/mz to a tone.
            """
            try:
                mz = float(mz)
            except ValueError:
                return 100
            wv = (mz * (max_hz - min_hz) -
                  max_hz * min_mz + min_hz * max_mz) / (max_mz - min_mz)
            return int(44100 / wv)

        # go through each trace and map it into the sound array
        for i, mz in enumerate(self.columns):
            if float(mz) < cutoff:
                # clip out mz/wv below a certain threshold
                # handy if data has low level noise
                continue
            print(str(i) + '/' + str(self.shape[1]))
            inter_x = np.linspace(0, 1, wav[::mz_to_wv(mz)].shape[0])
            wav[::mz_to_wv(mz)] += np.interp(inter_x, tmask, self.values[:, i])

        # scale the new array and write it out
        scaled = wav / np.max(np.abs(wav))
        scaled = scipy.signal.fftconvolve(scaled, np.ones(5) / 5, mode='same')
        scaled = np.int16(scaled * 32767)
        scipy.io.wavfile.write(filename, 44100, scaled)

    def scans(self):
        for t in self.index:
            yield self.scan(t)

    def scan(self, t, dt=None, aggfunc=None):
        """
        Returns the spectrum from a specific time.

        Parameters
        ----------
        t : float
        dt : float
        """
        idx = (np.abs(self.index - t)).argmin()

        if dt is None:
            # only take the spectra at the nearest time
            mz_abn = self.values[idx, :].copy()
        else:
            # sum up all the spectra over a range
            en_idx = (np.abs(self.index - t - dt)).argmin()
            idx, en_idx = min(idx, en_idx), max(idx, en_idx)
            if aggfunc is None:
                mz_abn = self.values[idx:en_idx + 1, :].copy().sum(axis=0)
            else:
                mz_abn = aggfunc(self.values[idx:en_idx + 1, :].copy())
        if isinstance(mz_abn, scipy.sparse.spmatrix):
            mz_abn = mz_abn.toarray()[0]
        return Scan(self.columns, mz_abn)

    def compress(self):
        """
        Serializes the AstonFrame into a binary stream.

        Returns
        -------
        bytes
        """
        i = self.index.astype(np.float32).tostring()
        li = struct.pack('<L', len(i))
        c = json.dumps(self.columns).encode('utf-8')
        lc = struct.pack('<L', len(c))
        v = self.values.astype(np.float64).tostring()
        try:  # python 2
            return buffer(zlib.compress(lc + li + c + i + v))
        except NameError:  # python 3
            return zlib.compress(lc + li + c + i + v)


def decompress(zdata):
    """
    Unserializes an AstonFrame.

    Parameters
    ----------
    zdata : bytes

    Returns
    -------
    Trace or Chromatogram

    """
    data = zlib.decompress(zdata)
    lc = struct.unpack('<L', data[0:4])[0]
    li = struct.unpack('<L', data[4:8])[0]
    c = json.loads(data[8:8 + lc].decode('utf-8'))
    i = np.frombuffer(data[8 + lc:8 + lc + li], dtype=np.float32)
    v = np.frombuffer(data[8 + lc + li:], dtype=np.float64)

    if len(c) == 1:
        return Trace(v, i, name=c[0])
    else:
        return Chromatogram(v.reshape(len(i), len(c)), i, c)


def _slice_idxs(df, twin=None):
    """
    Returns a slice of the incoming array filtered between
    the two times specified. Assumes the array is the same
    length as self.data. Acts in the time() and trace() functions.
    """
    if twin is None:
        return 0, df.shape[0]

    tme = df.index

    if twin[0] is None:
        st_idx = 0
    else:
        st_idx = (np.abs(tme - twin[0])).argmin()
    if twin[1] is None:
        en_idx = df.shape[0]
    else:
        en_idx = (np.abs(tme - twin[1])).argmin() + 1
    return st_idx, en_idx



# this is used here and in the peak plotting code
def desaturate(c, k=0):
    """
    Utility function to desaturate a color c by an amount k.
    """
    from matplotlib.colors import ColorConverter
    c = ColorConverter().to_rgb(c)
    intensity = 0.299 * c[0] + 0.587 * c[1] + 0.114 * c[2]
    return [intensity * k + i * (1 - k) for i in c]


def plot_events(events, color='k', ax=None):
    from matplotlib.transforms import offset_copy
    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()

    trans = ax.get_xaxis_transform()
    trans_text = offset_copy(trans, fig=ax.figure, x=3, units='points')
    for ev in events:
        t0, t1, ta = ev['t0'], ev['t1'], (ev['t0'] + ev['t1']) / 2.
        ax.vlines(t1, 0, 0.1, color=desaturate(color, 0.6), transform=trans)
        ax.vlines(t0, 0, 0.1, color=color, transform=trans)
        ax.text(ta, 0, ev['name'], ha='center', transform=trans_text)



import numpy as np
from scipy import convolve
from aston.spectra import Scan
from aston.trace.trace import Trace
# from aston.spectra.isotopes import delta13c_santrock, delta13c_craig
from aston.peak.peak_fitting import fit, guess_initc
from aston.peak.peak_models import peak_models
from aston.trace.events import desaturate

peak_models = dict([(pm.__name__, pm) for pm in peak_models])


class Peak(object):
    def __init__(self, name='', info=None, components=None):
        self.name = name
        if info is None:
            self.info = {}
        else:
            self.info = info

        if isinstance(components, (list, tuple)):
            self.components = list(components)
        elif isinstance(components, PeakComponent):
            self.components = [components]
        else:
            self.components = []

    def as_poly(self, mz=None, sub_base=False):
        # TODO: should allow Chromatogram in PeakComponents some day?
        # if type(self.trace) is Trace:
        #     if mz is not None:
        #         assert mz == self.trace.name
        #     trace = self.trace
        # else:
        #     if mz is None:
        #         mz = self.primary_mz
        #     elif mz not in self.trace.columns:
        #         return Trace()
        #     trace = self.trace[mz]

        # if self.baseline is None:
        #     b_trace = None
        # elif type(self.baseline) is Trace:
        #     b_trace = self.baseline
        # else:
        #     b_trace = self.baseline[mz]

        if mz in {'', 'x', 'tic', None}:
            # sum up all the components
            trace = self.components[0].trace
            b_trace = self.components[0].baseline
            for c in self.components[1:]:
                trace += c.trace
                if c.baseline is not None:
                    # TODO: this fails if the first components baseline is None
                    b_trace += c.baseline
        else:
            tol = 0.5

            def check(name):
                try:
                    return np.abs(mz - float(name)) < tol
                except ValueError:
                    return False

            cs = [c for c in self.components if check(c._trace.name)]
            trace = sum(c.trace for c in cs)
            b_trace = sum(c.baseline for c in cs)

        # merge the trace and baseline
        if sub_base and b_trace is not None:
            trace -= np.interp(trace.index, b_trace.index,
                               b_trace.values)
            t, z = trace.index, trace.values
        elif b_trace is None:
            t, z = trace.index, trace.values
        else:
            t = np.hstack([trace.index, b_trace.index[::-1]])
            z = np.hstack([trace.values, b_trace.values[::-1]])

        if hasattr(self, 'dbplot'):
            # scale and offset according to parent
            t *= self.dbplot.x_scale
            t += self.dbplot.x_offset
            z *= self.dbplot.y_scale
            z += self.dbplot.y_offset

        return np.vstack([t, z]).T

    def plot(self, mz=None, color='k', alpha=0.5, ax=None):
        from matplotlib.path import Path
        from matplotlib.patches import PathPatch

        # TODO: allow plotting to a 2D plot?
        # confine matplotlib imports to here, so this module works
        # even if matplotlib is not installed (e.g. on a server)
        path = self.as_poly(mz)
        ply = PathPatch(Path(path), facecolor=desaturate(color, 0.5),
                        alpha=alpha, lw=0)
        if ax is None:
            import matplotlib.pyplot as plt
            plt.add_patch(ply)
        else:
            ax.add_patch(ply)

    def contains(self, x, y, mz=None):
        from matplotlib.path import Path
        return Path(self.as_poly(mz)).contains_point((x, y))

    def time(self, mz=None):
        # TODO: determine inversion by comparing baseline to trace?
        data = self.as_poly(mz)
        if data[1, 0] < data[:, 0].max():
            return data[data[:, 1].argmax(), 0]
        else:  # inverted peak
            return data[data[:, 1].argmin(), 0]

    def pwhm(self, mz=None):
        data = self.as_poly(mz)
        pt1, pt2 = data[0], data[-1]

        m = (pt2[1] - pt1[1]) / (pt2[0] - pt1[0])
        avs = np.array([(pt[0], (pt[1] - m * (pt[0] - pt1[0]) - pt1[1]))
                        for pt in data])

        # calculate the height of half-max
        half_y = max(avs[:, 1]) / 2.0
        lw_x, hi_x = float('nan'), float('nan')
        # loop through all of the line segments
        for i in range(len(avs) - 1):
            # does this line segment intersect half-max?
            if (avs[i, 1] < half_y and avs[i + 1, 1] > half_y) or \
                    (avs[i, 1] > half_y and avs[i + 1, 1] < half_y):
                m = (avs[i + 1, 1] - avs[i, 1]) / (avs[i + 1, 0] - avs[i, 0])
                b = (avs[i + 1, 0] * avs[i, 1] - avs[i, 0] * avs[i + 1, 1])
                b /= (avs[i + 1, 0] - avs[i, 0])
                if np.isnan(lw_x) and np.isnan(hi_x):
                    lw_x, hi_x = (half_y - b) / m, (half_y - b) / m
                else:
                    lw_x = min((half_y - b) / m, lw_x)
                    hi_x = max((half_y - b) / m, hi_x)
        return hi_x - lw_x

    def height(self, mz=None):
        data = self.as_poly(mz, sub_base=True)
        return data[:, 1].max() - data[:, 1].min()

    def width(self, mz=None):
        data = self.as_poly(mz)
        return data[:, 0].max() - data[:, 0].min()

    def area(self, mz=None, method='shoelace'):
        data = self.as_poly(mz)

        # filter out any points that have a nan
        fdata = data[~np.isnan(data).any(1)]

        if method == 'shoelace':
            # up to 5e-10 diff from shoelace-slow
            csum = np.sum(np.fliplr(np.roll(fdata, 1, axis=0)) * fdata, axis=0)
            return 0.5 * np.abs(csum[0] - csum[1])
        elif method == 'shoelace-slow':
            csum = 0
            x, y = fdata[-1, :]
            for i in fdata:
                csum += i[0] * y - i[1] * x
                x, y = i
            return abs(csum / 2.)
        elif method == 'trapezoid':
            # http://en.wikipedia.org/wiki/trapezoidal_rule#non-uniform_grid
            # todo: this essentially ignores baseline data?
            # fdata[:, 1][fdata[:, 1] < 0] = 0
            # y = convolve(fdata[:, 1], [0.5, 0.5], mode='valid')

            # y = convolve(np.abs(fdata[:, 1]), [0.5, 0.5], mode='valid')

            y = convolve(fdata[:, 1], [0.5, 0.5], mode='valid')
            if y.shape[0] != fdata.shape[0] - 1:
                return 0
            return np.sum(np.diff(fdata[:, 0]) * y)
        elif method == 'sum':
            return np.sum(fdata[:, 1])

    def as_scan(self, aggtype=None):
        return Scan()


class PeakComponent(object):
    def __init__(self, info, trace=None, baseline=None):
        self._trace = trace
        self.baseline = baseline
        self.info = info

    # this allows transparent switching to/from analytical peaks
    @property
    def trace(self):
        if self.info.get('p-model') not in peak_models:
            return self._trace
        else:
            model = peak_models[self.info.get('p-model')]
            t = self._trace.index
            d = model(t=t, **{k: self.info.get(k)
                              for k in model._peakargs
                              if self.info.get(k) is not None})
            d += self.baseline._retime(t)
            return Trace(d, t, name='')

    def refit(self, peak_model=None):
        # TODO: needs to work with PeakComponents containing DataFrames?
        self.info['p-model'] = peak_model
        if peak_model is None:
            # TODO: remove parameters from self.info?
            # del self.info['p-model-fit']
            return
        model = peak_models[peak_model]

        # remove the baseline and create a new series
        t = self._trace.index
        d = self._trace.values - self.baseline._retime(t)
        ts = Trace(d, t)

        # fit the peak
        initc = guess_initc(ts, model, [t[d.argmax()]])
        params, res = fit(ts, [model], initc)
        self.info.update(params[0])
        self.info['p-model-fit'] = res['r^2']

    def time(self):
        if np.average(self.trace.values) > np.average(self.baseline.values):
            return self.trace.index[self.trace.values.argmax()]
        else:
            return self.trace.index[self.trace.values.argmin()]

# class OldPeak(object):
#     def __init__(self, *args, **kwargs):
#         super(Peak, self).__init__(*args, **kwargs)
#         self.db_type = 'peak'
#         self.childtypes = ('spectrum',)
#
#     @property
#     def data(self):
#         if 'p-model' not in self.info:
#             return self.rawdata
#
#         #TODO: move this into the loop, so each mz
#         # can theoretically have its own function-type
#         f = peak_models.get(self.info['p-model'], None)
#         if f is None:
#             return self.rawdata
#
#         all_params = json.loads(self.info['p-params'])
#         y = np.empty(self.rawdata.data.shape)
#         for i, params in enumerate(all_params):
#             y[:, i] = f(self.rawdata.times, **params)
#             y[:, i] += self.baseline(params['mz'], True)
#         return TimeSeries(y, self.rawdata.times, self.rawdata.ions)
#
#     def baseline(self, ion=None, interp=False):
#         if self.info['p-baseline'] == '':
#             return None
#         bases = json.loads(self.info['p-baseline'])
#         new_bases = bases.copy()
#         #TODO: incredibly hacky and slow
#         for b in bases:
#             try:
#                 new_bases[float(b)] = bases[b]
#             except:
#                 pass
#         if ion in new_bases:
#             if interp:
#                 base_pts = np.array(new_bases[ion]).T
#                 return np.interp(self.rawdata.times, *base_pts)
#             else:
#                 return np.array(new_bases[ion])
#         else:
#             return None
#
#     def set_baseline(self, ion, value=None):
#         if self.info['p-baseline'] == '':
#             bases = {}
#         else:
#             bases = json.loads(self.info['p-baseline'])
#         if value is None and str(ion) in bases:
#             del bases[str(ion)]
#         elif value is None:
#             return
#         else:
#             assert type(value) == np.ndarray
#         bases[str(ion)] = value.tolist()
#         self.info['p-baseline'] = json.dumps(bases)
#
#     def time(self, twin=None):
#         return self.rawdata.trace('!', twin=twin).time
#
#     def trace(self, ion='!', twin=None):
#         return self.rawdata.trace(ion, twin=twin).y
#
#     def _load_info(self, fld):
#         if fld == 's-mzs':
#             ions = self.data.ions
#             if len(ions) < 10:
#                 self.info[fld] = ','.join(str(i) for i in ions)
#             else:
#                 # only display a range of the numeric ions
#                 ions = [i for i in ions \
#                   if type(i) is int or type(i) is float]
#                 if len(ions) > 0:
#                     self.info['s-mzs'] = str(min(ions)) + '-'
#                     self.info['s-mzs'] += str(max(ions))
#         elif fld == 'p-s-area':
#             self.info[fld] = str(self.area('!'))
#         elif fld == 'p-s-length':
#             self.info[fld] = str(peakmath.length(self.as_poly()))
#         elif fld == 'p-s-height':
#             self.info[fld] = str(peakmath.height(self.as_poly()))
#         elif fld == 'p-s-time':
#             self.info[fld] = str(peakmath.time(self.as_poly()))
#         elif fld == 'p-s-pwhm':
#             self.info[fld] = str(peakmath.length(self.as_poly(), pwhm=True))
#         elif fld == 'r-filename':
#             prt = self.parent_of_type('file')
#             self.info[fld] = prt.info['r-filename']
#
#     def _calc_info(self, fld):
#         if fld == 'p-s-pkcap':
#             prt = self.parent_of_type('file')
#             if prt is None:
#                 return ''
#             t = float(prt.info['s-peaks-en']) - \
#                 float(prt.info['s-peaks-st'])
#             return str(t / peakmath.length(self.as_poly()) + 1)
#         elif fld == 'p-s-d13c':
#             return self.d13c()
#         return ''
#
#     def contains(self, x, y, ion=None):
#         if not self.data.has_ion(ion):
#             return False
#         return peakmath.contains(self.as_poly(ion), x, y)
#
#     def as_poly(self, ion=None, sub_base=False):
#         # add in the baseline on either side
#         if ion is None:
#             row = 0
#             ion = self.data.ions[0]
#         elif not self.data.has_ion(ion):
#             row = 0
#             ion = self.data.ions[0]
#         else:
#             try:
#                 row = self.data.ions.index(float(ion))
#             except ValueError:
#                 row = self.data.ions.index(ion)
#         pk = np.vstack([self.data.times, self.data.data.T[row]]).T
#         base = self.baseline(ion)
#         if sub_base:
#             # this subtracts out the base line before returning it
#             # it's useful for numerical fxns that don't take baseline
#             if base is None:
#                 base_pts = np.interp(pk[:, 0], [pk[1, 0], pk[-1, 0]], \
#                                      [pk[0, 1], pk[-1, 1]])
#             else:
#                 base_pts = np.interp(pk[:, 0], *base)
#
#             ply = np.array([pk[:, 0], pk[:, 1] - base_pts]).T
#         elif base is None:
#             ply = pk
#         else:
#             ply = np.vstack([base[0], pk, base[:0:-1]])
#         return ply[np.logical_not(np.any(np.isnan(ply), axis=1))]
#
#     def area(self, ion=None):
#         if ion == '!':
#             pk = self.as_poly()  # sub_base=True)
#         elif not self.data.has_ion(ion):
#             return 0
#         else:
#             pk = self.as_poly(ion)  # , sub_base=True)
#         #if peakmath.area(pk, method='shoelace') / \
#         #   peakmath.area(pk, method='trapezoid') != 1:
#         return peakmath.area(pk)
#
#     def d13c(self):
#         dt = self.parent_of_type('file')
#         #TODO: not sure if we should do this or not
#         # by not doing it, we can show relative error
#         # between standard peaks
#         #if self.info['p-type'] == 'Isotope Standard':
#         #    return dt.info['r-d13c-std']
#
#         # if there's no reference number, we can't do this
#         try:
#             float(dt.info['r-d13c-std'])
#         except:
#             return ''
#
#         if self.db is not None:
#             calc_meth = self.db.get_key('d13c_method', dflt='santrock')
#             consts = self.db.get_key('d13c_const', dflt='Santrock')
#         else:
#             calc_meth, consts = 'santrock', 'Santrock'
#
#         r45std = dt.get_point('r45std', peakmath.time(self.as_poly(44)))
#         r46std = dt.get_point('r46std', peakmath.time(self.as_poly(44)))
#
#         # if no peak has been designated as a isotope std
#         if r45std == 0.0:
#             return ''
#
#         i44, i45, i46 = self.area(44), self.area(45), self.area(46)
#         # if one of the areas is 0, clearly there's a problem
#         if i44 * i45 * i46 == 0:
#             return ''
#         if calc_meth == 'craig':
#             d = delta13c_craig(i45 / i44, i46 / i44, \
#               float(dt.info['r-d13c-std']), r45std, r46std)
#         else:
#             d = delta13c_santrock(i45 / i44, i46 / i44, \
#               float(dt.info['r-d13c-std']), r45std, r46std,
#               ks=consts)
#
#         return '{0:.3f}'.format(d)
#
#     def as_spectrum(self, method=None):
#         if method is None:
#             # grab the spectrum from the same time as me
#             prt = self.parent_of_type('file')
#             time = peakmath.time(self.as_poly())
#             data = prt.scan(time)
#         elif method == 'child':
#             # if there's already a sprectum assigned to
#             # me, return it
#             specs = self.children_of_type('spectrum')
#             if len(specs) > 0:
#                 return specs[0]
#         return Spectrum({'p-s-time': str(time)}, data)
#
#     def update_model(self, new_f):
#         INITC_ONLY = False
#         self.info['p-model'] = str(new_f)
#         self.info.del_items('p-s-')
#
#         f = peak_models.get(str(new_f), None)
#         all_params = []
#         if f is not None:
#             for i in self.rawdata.ions:
#                 t = self.rawdata.times
#                 y = self.rawdata.trace(i).y
#                 #TODO: subtract baseline
#                 #ya = x[1:-1] - np.linspace(x[0], x[-1], len(x) - 2)
#                 y -= self.baseline(i, interp=True)
#
#                 ts = TimeSeries(y, t)
#                 initc = guess_initc(ts, f, [t[y.argmax()]])
#                 if INITC_ONLY:
#                     params = initc[0]
#                 else:
#                     params, res = fit(ts, [f], initc)
#                     params = params[0]
#
#                 params['mz'] = str(i)
#                 params['f'] = str(new_f)
#                 if not INITC_ONLY:
#                     params['r^2'] = str(res['r^2'])
#                 all_params.append(params)
#
#             self.info['p-params'] = json.dumps(all_params)
#
#             get_param = lambda k, l: ','.join(str(p[k]) for p in l)
#             self.info['p-s-height'] = str(get_param('h', all_params))
#             #self.info['p-s-base'] = str(get_param('v', all_params))
#             self.info['p-s-time'] = str(get_param('x', all_params))
#             self.info['p-s-width'] = str(get_param('w', all_params))
#             if not INITC_ONLY:
#                 self.info['p-s-model-fit'] = str(get_param('r^2',
#                                                            all_params))



import os
import re
import struct
from datetime import datetime
import numpy as np
from aston.cache import cache
from aston.trace import Chromatogram
from aston.tracefile import TraceFile


def string_read(f):
    """modified string read method which works with our UV files"""
    # todo figure out why there is a null character between every value
    #   - check if this is true across all cases
    #   - if so, does this affect the interpretation of intensity values
    # it's not pretty, but it works
    read_len = struct.unpack(  # determine length to read
        '>B', f.read(1)
    )[0]
    out = f.read(  # read values, decode, and strip
        2 * read_len - 1
    ).decode('ascii').strip()
    return out.replace('\x00', '')


class AgilentMWD(TraceFile):
    mime = 'application/vnd-agilent-chemstation-mwd'
    traces = ['#uv']

    @property
    def data(self):
        print(self.filename)
        # Because the spectra are stored in several files in the same
        # directory, we need to loop through them and return them together.
        ions = []
        dtraces = []
        foldname = os.path.dirname(self.filename)
        # if foldname == '': foldname = os.curdir
        for i in [os.path.join(foldname, i) for i in
                  os.listdir(foldname)]:
            if i[-3:].upper() == '.CH':
                wv, dtrace = self._read_ind_file(i)

                if wv is None:
                    continue

                # generate the time points if this is the first trace
                if len(ions) == 0:
                    f = open(i, 'rb')
                    f.seek(0x244)
                    yunits = f.read(struct.unpack('>B', f.read(1))[0]).decode()
                    f.seek(0x11A)
                    st_t = struct.unpack('>i', f.read(4))[0] / 60000.
                    en_t = struct.unpack('>i', f.read(4))[0] / 60000.
                    f.close()

                    times = np.linspace(st_t, en_t, len(dtrace))

                # add the wavelength and trace into the appropriate places
                ions.append(float(wv))
                dtraces.append(dtrace)
        data = np.array(dtraces).transpose()
        return Chromatogram(data, times, ions, yunits=yunits)

    @staticmethod
    def _read_ind_file(fname):
        f = open(fname, 'rb')

        if f.read(2) != b'\x02\x33':
            return None, None

        f.seek(0x254)
        sig_name = str(f.read(struct.unpack('>B', f.read(1))[0]))
        # wavelength the file was collected at
        wv = re.search('[\\w]+=(\\d+)', sig_name).group(1)

        f.seek(0x284)
        del_ab = struct.unpack('>d', f.read(8))[0]

        data = np.array([])

        f.seek(0x401)
        loc = 0
        while True:
            rec_len = struct.unpack('>B', f.read(1))[0]
            if rec_len == 0:
                break
            data = np.append(data, np.empty(rec_len))
            for _ in range(rec_len):
                inp = struct.unpack('>h', f.read(2))[0]
                if inp == -32768:
                    inp = struct.unpack('>i', f.read(4))[0]
                    data[loc] = del_ab * inp
                elif loc == 0:
                    data[loc] = del_ab * inp
                else:
                    data[loc] = data[loc - 1] + del_ab * inp
                loc += 1
            f.read(1)  # this value is always 0x10?
        f.close()
        return wv, data

    @property
    def info(self):
        d = super(AgilentMWD, self).info
        # TODO: fix this so that it doesn't rely upon MWD1A.CH?
        f = open(self.filename, 'rb')
        f.seek(0x18)
        d['name'] = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        f.seek(0x94)
        d['r-opr'] = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        f.seek(0xE4)
        d['m-name'] = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        # try:
        f.seek(0xB2)
        rawdate = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        d['r-date'] = datetime.strptime(rawdate,
                                        '%d-%b-%y, %H:%M:%S').isoformat(' ')
        # except: pass #TODO: find out why this chokes
        f.seek(0x254)
        # TODO: replace signal name with reference_wavelength?
        # d['signal name'] = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        f.close()
        return d


class AgilentMWD2(TraceFile):
    mime = 'application/vnd-agilent-chemstation-mwd2'
    traces = ['#uv']

    @property
    def data(self):
        # Because the spectra are stored in several files in the same
        # directory, we need to loop through them and return them together.
        ions = []
        dtraces = []
        foldname = os.path.dirname(self.filename)
        for i in [os.path.join(foldname, i) for i in
                  os.listdir(foldname)]:
            if i[-3:].upper() == '.CH':
                wv, dtrace = self._read_ind_file(i)

                # generate the time points if this is the first trace
                if len(ions) == 0:
                    f = open(i, 'rb')
                    yunits = self._get_str(f, 0x104C)
                    f.seek(0x11A)
                    st_t = struct.unpack('>i', f.read(4))[0] / 60000.
                    en_t = struct.unpack('>i', f.read(4))[0] / 60000.
                    f.close()

                    times = np.linspace(st_t, en_t, len(dtrace))

                # add the wavelength and trace into the appropriate places
                ions.append(float(wv))
                dtraces.append(dtrace)
        data = np.array(dtraces).transpose()
        return Chromatogram(data, times, ions, yunits=yunits)

    def _read_ind_file(self, fname):
        f = open(fname, 'rb')

        f.seek(0x1075)
        sig_name = f.read(2 * struct.unpack('>B',
                                            f.read(1))[0]).decode('utf-16')
        # wavelength the file was collected at
        wv = re.search('[\\w]+=(\\d+)', sig_name).group(1)

        f.seek(0x127C)
        del_ab = struct.unpack('>d', f.read(8))[0]

        data = []

        f.seek(0x1800)
        while True:
            x, nrecs = struct.unpack('>BB', f.read(2))
            if x == 0 and nrecs == 0:
                break
            for _ in range(nrecs):
                inp = struct.unpack('>h', f.read(2))[0]
                if inp == -32768:
                    inp = struct.unpack('>i', f.read(4))[0]
                    data.append(del_ab * inp)
                elif len(data) == 0:
                    data.append(del_ab * inp)
                else:
                    data.append(data[-1] + del_ab * inp)
        f.close()
        return wv, np.array(data)

    def _get_str(self, f, off):
        """
        Convenience function to quickly pull out strings.
        """
        f.seek(off)
        return f.read(2 * struct.unpack('>B', f.read(1))[0]).decode('utf-16')

    @property
    def info(self):
        d = super(AgilentMWD2, self).info
        # TODO: fix this so that it doesn't rely upon MWD1A.CH?

        f = open(self.filename, 'rb')
        d['name'] = self._get_str(f, 0x35A)
        d['r-opr'] = self._get_str(f, 0x758)
        d['m-name'] = self._get_str(f, 0xA0E)
        rawdate = self._get_str(f, 0x957)
        try:
            d['r-date'] = datetime.strptime(rawdate, '%d-%b-%y, %H:%M:%S').isoformat(' ')  # noqa
        except ValueError:
            try:
                d['r-date'] = datetime.strptime(rawdate, '%d %b %y  %I:%M %p').isoformat(' ')  # noqa
            except ValueError:
                pass
        # TODO: replace signal name with reference_wavelength?
        # d['signal name'] = f.read(struct.unpack('>B', f.read(1))[0]).decode()
        f.close()
        return d


class AgilentDAD(TraceFile):
    mime = 'application/vnd-agilent-masshunter-dad'
    traces = ['#uv']

    # header data in DAD1.sd
    # 80 byte repetition
    # offset = 0xA4, format = 'IIfIfffQIIdddd'
    #  750 entries
    # f.seek(0xA4); struct.unpack('<IddffIQIIdddd', f.read(80))

    # time series in DAD1.sg
    # all doubles, starts at 0x44
    # 750x 54 double entries
    @property
    def data(self):
        fhead = open(self.filename[:-3] + '.sd', 'rb')
        fdata = open(self.filename[:-3] + '.sp', 'rb')

        fhead.seek(0x50)
        nscans = struct.unpack('Q', fhead.read(8))[0]
        fhead.seek(0xA4)

        ions = []
        for scn in range(nscans):
            t = struct.unpack('<IdddIQIIdddd', fhead.read(80))

            npts = t[7]
            if scn == 0:
                # TODO: this will fail if the wavelengths collected
                # change through the run.
                data = np.zeros((nscans, npts))
                times = np.zeros((nscans))
                ions = [t[8] + x * t[3] for x in range(npts)]

            times[scn] = t[1]

            fdata.seek(t[5] + 16)
            # TODO: use np.fromfile?
            data[scn] = struct.unpack('<' + npts * 'd', fdata.read(npts * 8))
        return Chromatogram(data, times, ions)

        fhead.close()
        fdata.close()


class AgilentCSDAD(TraceFile):
    """
    Interpreter for *.UV files from Agilent Chemstation
    """
    mime = 'application/vnd-agilent-chemstation-dad'
    traces = ['#uv']

    @property
    @cache(maxsize=1)
    def data(self):
        # TODO: the chromatograms this generates are not exactly the
        # same as the ones in the *.CH files. Maybe they need to be 0'd?
        f = open(self.filename, 'rb')

        f.seek(0x146)
        yunits = f.read(struct.unpack('>B',
                                      f.read(1))[0]).decode('ascii').strip()

        f.seek(0x116)
        nscans = struct.unpack('>i', f.read(4))[0]

        times = np.zeros(nscans)
        data = nscans * [{}]
        ions = []
        npos = 0x202
        for i in range(nscans):
            f.seek(npos)
            npos += struct.unpack('<H', f.read(2))[0]
            times[i] = struct.unpack('<L', f.read(4))[0] / 60000.
            nm_srt = struct.unpack('<H', f.read(2))[0] / 20.
            nm_end = struct.unpack('<H', f.read(2))[0] / 20.
            nm_stp = struct.unpack('<H', f.read(2))[0] / 20.
            f.read(8)
            s = {}
            v = struct.unpack('<h', f.read(2))[0] / 2000.
            s[nm_srt] = v
            for wv in np.arange(nm_srt, nm_end, nm_stp):
                ov = struct.unpack('<h', f.read(2))[0]
                if ov == -32768:
                    v = struct.unpack('<i', f.read(4))[0] / 2000.
                else:
                    v += ov / 2000.
                s[wv] = v
                if wv not in ions:
                    ions.append(wv)
            data[i] = s

        ndata = np.zeros((nscans, len(ions)))
        for i, d in zip(range(nscans), data):
            for ion, abn in d.items():
                ndata[i, ions.index(ion)] = abn
        return Chromatogram(ndata, times, ions, yunits=yunits)

    @property
    def info(self):
        d = super(AgilentCSDAD, self).info
        with open(self.filename, 'rb') as f:
            f.seek(0x18)
            d['name'] = string_read(f)
            f.seek(0x94)
            d['r-opr'] = string_read(f)
            f.seek(0xE4)
            d['m-name'] = string_read(f)
            f.seek(0xB2)
            rawdate = string_read(f)
            try:  # fails on 0331 UV files
                d['r-date'] = datetime.strptime(rawdate, "%d-%b-%y, %H:%M:%S").isoformat(' ')  # noqa
            except TypeError:
                pass
            f.seek(0xD0)
            d['r-inst'] = string_read(f)
            # TODO: are the next values correct?
            f.seek(0xFE)
            d['r-vial-pos'] = string_read(f)
            f.seek(0xFE)
            d['r-seq-num'] = string_read(f)
        return d


class AgilentCSDAD2(TraceFile):
    """
    Interpreter for *.UV files from Agilent Chemstation
    """
    mime = 'application/vnd-agilent-chemstation-dad2'
    traces = ['#uv']

    @property
    @cache(maxsize=1)
    def data(self):
        f = open(self.filename, 'rb')

        f.seek(0xC15)
        yunits = string_read(f)

        f.seek(0x116)
        nscans = struct.unpack('>i', f.read(4))[0]

        # get all wavelengths and times
        wvs = set()
        times = np.empty(nscans)
        npos = 0x1002
        for i in range(nscans):
            f.seek(npos)
            npos += struct.unpack('<H', f.read(2))[0]
            times[i] = struct.unpack('<L', f.read(4))[0]
            nm_srt, nm_end, nm_stp = struct.unpack('<HHH', f.read(6))
            n_wvs = np.arange(nm_srt, nm_end, nm_stp) / 20.
            wvs.update(set(n_wvs).difference(wvs))
        wvs = list(wvs)

        ndata = np.empty((nscans, len(wvs)), dtype="<i4")
        npos = 0x1002

        # try to speed up by preloading the function
        unpack = struct.unpack
        seek = f.seek
        read = f.read
        tell = f.tell

        for i in range(nscans):
            seek(npos)
            dlen = unpack('<H', read(2))[0]
            npos += dlen
            seek(tell() + 4)  # skip time
            nm_srt, nm_end, nm_stp = unpack('<HHH', read(6))
            seek(tell() + 8)

            # OLD CODE
            v = 0
            pos = f.tell()
            for wv in np.arange(nm_srt, nm_end, nm_stp) / 20.:
                ov = unpack('<h', read(2))[0]
                if ov == -32768:
                    v = unpack('<i', read(4))[0]
                else:
                    v += ov
                ndata[i, wvs.index(wv)] = v
            seek(pos)

            #  WORKING ON A FASTER WAY TO READ ALL THIS DATA BELOW
            #  read in all the data
            # data = np.fromfile(f, dtype="<i2", count=int((dlen - 24) / 2))

            #  if there are any records marked -32768, we need to reinterpret
            #  parts of the array as i4's
            # oob_idxs = np.where(data == -32768)[0]

            #  locations of the cells to merge into 32-bit ints
            # big_idxs = np.repeat(oob_idxs, 2) + \
            #         np.tile([1, 2], oob_idxs.shape[0])
            # big_data = data[big_idxs].view('<i4').copy()

            #  remove the 32-bit cells, so the arrays the right size
            # data = np.delete(data, big_idxs).astype("<i4")

            #  set the marker cells to the right values
            # oob_idxs = np.where(data == -32768)[0]
            # data[oob_idxs] = big_data

            #  compute cumulative sums for each chunk
            # pidx = 0
            # if data.shape[0] != ndata.shape[1]:
            #     print(big_data)
            # for idx in np.hstack([oob_idxs, data.shape[0]]):
            #     ndata[i, pidx:idx] = np.cumsum(data[pidx:idx])
            #     pidx = idx
        f.close()
        return Chromatogram(ndata / 2000., times / 60000., wvs, yunits=yunits)

    @property
    def info(self):
        d = super(AgilentCSDAD2, self).info
        with open(self.filename, 'rb') as f:
            f.seek(0x35A)
            d['name'] = string_read(f)
            f.seek(0x758)
            d['r-opr'] = string_read(f)
            # get date into correct format before using this
            # f.seek(0x957)
            # d['r-date'] = string_read()
            f.seek(0xA0E)
            d['m-name'] = string_read(f)
        return d



import numpy as np
import scipy.sparse
from scipy.io.netcdf import NetCDFFile
from aston.trace.trace import Chromatogram, Trace
from aston.tracefile import TraceFile


class NetCDF(TraceFile):
    mime = 'application/netcdf'
    traces = ['#ms']

    def total_trace(self, twin=None):
        f = NetCDFFile(open(self.filename, 'rb'))
        tme = f.variables['scan_acquisition_time'].data / 60.
        tic = f.variables['total_intensity'].data
        return Trace(tic, tme, name='TIC')

    @property
    def data(self):
        f = NetCDFFile(open(self.filename, 'rb'))
        t = f.variables['scan_acquisition_time'].data / 60.

        # this is half the speed of the following code
        # ions = list(set(f.variables['mass_values'].data))
        # cols = np.array([ions.index(i) for i in \
        #                  f.variables['mass_values'].data])

        # TODO: slow; there has to be a way to vectorize this more?
        ions = np.array(list(set(f.variables['mass_values'].data)))
        rcols = f.variables['mass_values'].data
        cols = np.empty(rcols.shape, dtype=int)
        for i, ion in enumerate(ions):
            cols[rcols == ion] = i

        vals = f.variables['intensity_values'].data
        rowst = np.add.accumulate(f.variables['point_count'].data)
        rowst = np.insert(rowst, 0, 0)

        data = scipy.sparse.csr_matrix((vals, cols, rowst),
                                       shape=(len(t), len(ions)), dtype=float)
        return Chromatogram(data.todense(), t, ions)


def write_netcdf(filename, df, info=None):
    # FIXME: still a lot of issues here
    if info is None:
        info = {}

    f = NetCDFFile(filename, 'w', version=1)

    f.createDimension('_2_byte_string', 2)
    f.createDimension('_4_byte_string', 4)
    f.createDimension('_8_byte_string', 8)
    f.createDimension('_16_byte_string', 16)
    f.createDimension('_32_byte_string', 32)
    f.createDimension('_64_byte_string', 64)
    f.createDimension('_128_byte_string', 128)
    f.createDimension('_255_byte_string', 255)
    f.createDimension('error_number', 1)
    f.flush()

    # TODO: check that these do anything
    f.createDimension('instrument_number', 1)
    f.createDimension('range', 2)
    f.flush()

    f.dataset_completeness = 'C1'  # TODO: save peaks too? ('C1+C2')
    f.aia_template_revision = '1.0'
    f.ms_template_revision = '1.0.1'
    f.netcdf_revision = '2.3.2'
    f.languages = 'English'

    f.administrative_comments = 'none'
    f.dataset_origin = 'none'
    f.dataset_owner = ' '
    f.dataset_date_time_stamp = ' '
    f.flush()

    f.company_method_name = ' '
    f.pre_experiment_program_name = ' '
    f.post_experiment_program_name = ' '
    f.source_file_reference = ' '

    f.experiment_title = info.get('name', ' ')
    f.operator_name = info.get('operator', ' ')
    # TODO: wrong format for injection_date_time_stamp
    # example: 20141027123030-0500
    f.injection_date_time_stamp = info.get('date', ' ')
    f.company_method_id = info.get('method', ' ')
    f.sample_name = info.get('sample', ' ')
    f.flush()

    f.sample_id_comments = 'none'
    f.sample_id = 'none'
    f.sample_name = 'none'
    f.sample_type = 'none'
    f.sample_injection_volume = 'none'
    f.sample_amount = 'none'
    f.flush()

    f.retention_unit = 'Seconds'

    # TODO: need to merge this back into access method
    if hasattr(df.values, 'todense'):
        scan_locs = (df.values != 0).todense()
    else:
        scan_locs = df.values != 0

    f.createVariable('error_log', 'c', ('error_number', '_64_byte_string'))

    f.createDimension('scan_number', len(df.index))
    v = f.createVariable('scan_acquisition_time', '>d', ('scan_number',))
    v[:] = 60. * df.index.astype('d')
    v = f.createVariable('total_intensity', '>d', ('scan_number',))
    v[:] = df.values.sum(axis=1).astype('d').flatten()
    v = f.createVariable('point_count', '>i', ('scan_number',))
    v[:] = np.sum(scan_locs, axis=1).astype('i').flatten()
    f.flush()

    f.createDimension('point_number', np.sum(scan_locs))
    stretch_t = np.resize(df.index, df.values.T.shape).T
    v = f.createVariable('mass_values', '>f', ('point_number',))
    v[:] = stretch_t[scan_locs]
    v = f.createVariable('intensity_values', '>f', ('point_number',))
    if hasattr(df.values, 'todense'):
        v[:] = df.values.todense()[scan_locs]
    else:
        v[:] = df.values[scan_locs]

    # TODO: check that these do anything
    # f.createVariable('time_values', 'd', ('point_number',))
    v = f.createVariable('resolution', 'd', ('scan_number',))
    v[:] = -9999
    f.createVariable('actual_scan_number', 'i', ('scan_number',))
    v[:] = -9999
    f.createVariable('scan_index', 'i', ('scan_number',))
    v[:] = np.cumsum(np.sum(scan_locs, axis=1).astype('i'))
    f.createVariable('mass_range_min', 'd', ('scan_number',))
    v[:] = np.min(stretch_t[scan_locs]) * np.ones(stretch_t.shape[0])
    v = f.createVariable('mass_range_max', 'd', ('scan_number',))
    v[:] = np.max(stretch_t[scan_locs]) * np.ones(stretch_t.shape[0])
    v = f.createVariable('a_d_sampling_rate', 'd', ('scan_number',))
    v[:] = -9999
    v = f.createVariable('a_d_coaddition_factor', 'h', ('scan_number',))
    v[:] = -9999
    v = f.createVariable('flag_count', 'i', ('scan_number',))
    v[:] = 0
    f.createVariable('inter_scan_time', 'd', ('scan_number',))
    v[0] = 0
    v[1:] = np.diff(df.index.astype('d'))
    f.createVariable('scan_duration', 'd', ('scan_number',))
    v[:] = 1
    v = f.createVariable('time_range_min', 'd', ('scan_number',))
    v[:] = -9999
    v = f.createVariable('time_range_max', 'd', ('scan_number',))
    v[:] = -9999

    inst_tup = ('instrument_number', '_32_byte_string')
    f.createVariable('instrument_serial_no', 'c', inst_tup)
    f.createVariable('instrument_fw_version', 'c', inst_tup)
    f.createVariable('instrument_app_version', 'c', inst_tup)
    f.createVariable('instrument_os_version', 'c', inst_tup)
    f.createVariable('instrument_sw_version', 'c', inst_tup)
    f.createVariable('instrument_comments', 'c', inst_tup)
    f.createVariable('instrument_model', 'c', inst_tup)
    f.createVariable('instrument_name', 'c', inst_tup)
    f.createVariable('instrument_id', 'c', inst_tup)
    f.createVariable('instrument_mfr', 'c', inst_tup)

    f.close()
